# validated: 2016-10-27 DS a7eca7d src/Storage.cpp src/Storage.h
'''
    This tries to stay compatible with ntcore's persistence mechanism,
    but if you go outside the realm of normal operations it may differ.
'''

import ast
import binascii
import base64
import re

from .constants import *
from .value import Value

from .support.compat import RawConfigParser, NoSectionError, PY2

import logging
logger = logging.getLogger('nt')


PERSISTENT_SECTION = 'NetworkTables Storage 3.0'

_key_bool = re.compile('boolean "(.+)"')
_key_double = re.compile('double "(.+)"')
_key_string = re.compile('string "(.+)"')
_key_raw = re.compile('raw "(.+)"')
_key_bool_array = re.compile('array boolean "(.+)"')
_key_double_array = re.compile('array double "(.+)"')
_key_string_array = re.compile('array string "(.+)"')

_value_string = re.compile(r'"((?:\\.|[^"\\])*)",?')

# TODO: these escape functions almost certainly don't deal with unicode
#       correctly

# TODO: strictly speaking, this isn't 100% compatible with ntcore... but

def _unescape_string(s):
    # shortcut if no escapes present
    if '\\' not in s:
        return s
    
    # let python do the hard work
    return ast.literal_eval('"%s"' % s)

# This is mostly what we want... unicode strings won't work properly though
_table = {i: chr(i) if i >= 32 and i < 127 else '\\x%02x' % i for i in range(256)}
_table[ord('"')] = '\\"'
_table[ord('\\')] = '\\\\'
_table[ord('\n')] = '\\n'
_table[ord('\t')] = '\\t'
_table[ord('\r')] = '\\r'

if PY2:
    _table = dict(map(lambda v: (v[0], unicode(v[1])), _table.items()))
    def _escape_string(s):
        return unicode(s).translate(_table)
else:
    def _escape_string(s):
        return s.translate(_table)

def load_entries(fp, filename):
    
    entries = []
    
    parser = RawConfigParser()
    parser.optionxform = str
    
    try:
        if hasattr(parser, 'read_file'):
            parser.read_file(fp, filename)
        else:
            parser.readfp(fp, filename)
    except IOError:
        raise
    except Exception as e:
        raise IOError('Error reading persistent file: %s' % e)
    
    try:
        items = parser.items(PERSISTENT_SECTION)
    except NoSectionError:
        raise IOError("Persistent section not found")

    value = None
    m = None
    
    for k, v in items:
        
        # Reduces code duplication
        if value:
            entries.append((_unescape_string(m.group(1)), value))
            
        value = None
        
        m = _key_bool.match(k)
        if m:
            if v == 'true':
                value = Value.makeBoolean(True)
            elif v == 'false':
                value = Value.makeBoolean(False)
            else:
                logger.warn("Unrecognized boolean value %r for %s", v, m.group(1))
            continue
        
        m = _key_double.match(k)
        if m:
            try:
                value = Value.makeDouble(float(v))
            except ValueError as e:
                logger.warn("Unrecognized double value %r for %s", v, m.group(1))
                
            continue
        
        m = _key_string.match(k)
        if m:
            mm = _value_string.match(v)
            
            if mm:
                value = Value.makeString(_unescape_string(mm.group(1)))
            elif PY2 and v == '':
                value = Value.makeString('')
            else:
                logger.warn("Unrecognized string value %r for %s", v, m.group(1))
            continue
        
        m = _key_raw.match(k)
        if m:
            try:
                if PY2:
                    v = base64.b64decode(v)
                else:
                    v = base64.b64decode(v, validate=True)
                value = Value.makeRaw(v)
            except binascii.Error:
                logger.warn("Unrecognized raw value %r for %s", v, m.group(1))
            continue
        
        m = _key_bool_array.match(k)
        if m:
            bools = []
            arr = v.strip().split(',')
            if arr != ['']:
                for vv in arr:
                    vv = vv.strip()
                    if vv == 'true':
                        bools.append(True)
                    elif vv == 'false':
                        bools.append(False)
                    else:
                        logger.warn("Unrecognized bool '%s' in bool array %s'", vv, m.group(1))
                        bools = None
                        break
                
            if bools is not None:
                value = Value.makeBooleanArray(bools)    
            continue
        
        m = _key_double_array.match(k)
        if m:
            doubles = []
            arr = v.strip().split(',')
            if arr != ['']:
                for vv in arr:
                    try:
                        doubles.append(float(vv))
                    except ValueError:
                        logger.warn("Unrecognized double '%s' in double array %s", vv, m.group(1))
                        doubles = None
                        break
                
            value = Value.makeDoubleArray(doubles)    
            continue
        
        m = _key_string_array.match(k)
        if m:
            # Technically, this will let invalid inputs in... but,
            # I don't really care. Feel free to fix it if you do.
            strings = [_unescape_string(vv) for vv in _value_string.findall(v)]
            value = Value.makeStringArray(strings)
            continue
        
        logger.warn("Unrecognized type '%s'", k)
        
    if value:
        entries.append((_unescape_string(m.group(1)), value))
    
    return entries


def save_entries(fp, entries):
    
    parser = RawConfigParser()
    parser.optionxform = str
    parser.add_section(PERSISTENT_SECTION)
    
    for name, value in entries:
        if not value:
            continue
        
        t = value.type
        v = value.value 
        
        if t == NT_BOOLEAN:
            name = 'boolean "%s"' % _escape_string(name)
            vrepr = 'true' if v else 'false'
        elif t == NT_DOUBLE:
            name = 'double "%s"' % _escape_string(name)
            vrepr = str(v)
        elif t == NT_STRING:
            name = 'string "%s"' % _escape_string(name)
            vrepr = '"%s"' % _escape_string(v)
        elif t == NT_RAW:
            name = 'raw "%s"' % _escape_string(name)
            vrepr = base64.b64encode(v).decode('ascii')
        elif t == NT_BOOLEAN_ARRAY:
            name = 'array boolean "%s"' % _escape_string(name)
            vrepr = ','.join(['true' if vv else 'false' for vv in v])
        elif t == NT_DOUBLE_ARRAY:
            name = 'array double "%s"' % _escape_string(name)
            vrepr = ','.join([str(vv) for vv in v])
        elif t == NT_STRING_ARRAY:
            name = 'array string "%s"' % _escape_string(name)
            vrepr = '","'.join([_escape_string(vv) for vv in v])
            if vrepr:
                vrepr = '"%s"' % vrepr
        else:
            continue
        
        parser.set(PERSISTENT_SECTION, name, vrepr)
    
    if PY2:
        parser.write(fp)
    else:
        parser.write(fp, space_around_delimiters=False)    



# novalidate

from collections import namedtuple

#: NetworkTables Entry Information
EntryInfo = namedtuple('EntryInfo', [
    # Entry name
    'name',

    # Entry type
    'type',

    # Entry flags
    'flags',

    # Timestamp of last change to entry (type or value).
    #'last_change',
])


#: NetworkTables Connection Information
ConnectionInfo = namedtuple('ConnectionInfo', [
    'remote_id',
    'remote_ip',
    'remote_port',
    'last_update',
    'protocol_version',
])


#: NetworkTables RPC Parameter Definition
RpcParamDef = namedtuple('RpcParamDef', [
    'name',
    'def_value',
])

#: NetworkTables RPC Result Definition
RpcResultDef = namedtuple('RpcResultDef', [
    'name',
    'type',
])

#: NetworkTables RPC Definition
RpcDefinition = namedtuple('RpcDefinition', [
    'version',
    'name',
    'params',
    'results',
])


#: NetworkTables RPC Call Data
RpcCallInfo = namedtuple('RpcCallInfo', [
    'rpc_id',
    'call_uid',
    'name',
    'params',
])


# validated: 2016-10-27 DS a7eca7d src/Storage.cpp src/Storage.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2015. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import os
import threading

from .message import Message
from .network_connection import NetworkConnection
from .persistence import load_entries, save_entries
from .structs import EntryInfo, ConnectionInfo
from .value import Value

from .support.compat import monotonic, file_replace
from .support.lists import ensure_id_exists

from .constants import (
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    kExecuteRpc,
    kRpcResponse,
    
    NT_UNASSIGNED,
    NT_PERSISTENT,
    
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS
)

import logging
logger = logging.getLogger('nt')



class _Entry(object):
    __slots__ = ['name', 'value', 'flags',
                 'id', 'seq_num', 'rpc_callback',
                 'rpc_call_uid']
    
    def __init__(self, name, value=None, flags=0, seq_num=0):
        # We redundantly store the name so that it's available when accessing the
        # raw Entry* via the ID map.
        self.name = name
        
        # The current value and flags.
        self.value = value
        self.flags = flags
        
        # Unique ID for self entry as used in network messages.  The value is
        # assigned by the server, on the client this is 0xffff until an
        # entry assignment is received back from the server.
        self.id = 0xffff
        
        # Sequence number for update resolution.
        self.seq_num = seq_num
        
        # RPC callback function.  Null if either not an RPC or if the RPC is
        # polled.
        self.rpc_callback = None
        
        # Last UID used when calling self RPC (primarily for client use).  This
        # is incremented for each call.
        self.rpc_call_uid = 0
        
    def isPersistent(self):
        return (self.flags & NT_PERSISTENT) != 0 
    
    def increment_seqnum(self):
        self.seq_num += 1
        self.seq_num &= 0xffff
        
    def isSeqNewerThan(self, other):
        seq_num = self.seq_num
        if other < seq_num:
            return (seq_num - other) < 32768
        elif other > seq_num:
            return (other - seq_num) > 32768
        else:
            return True
        
    def __repr__(self):
        return "<_Entry name='%s' value=%s flags=%s id=%s seq_num=%s rpc_callback=%s rpc_call_uid=%s" % \
            (self.name, self.value, self.flags, self.id,
             self.seq_num, self.rpc_callback, self.rpc_call_uid)


class Storage(object):
    
    def __init__(self, notifier, rpc_server):
        self.m_notifier = notifier
        self.m_rpc_server = rpc_server
        
        self.m_mutex = threading.Lock()
        self.m_entries = {}
        self.m_idmap = []
        self.m_rpc_results = {}
        self.m_rpc_blocking_calls = set()
        
        # If any persistent values have changed
        self.m_persistent_dirty = False
        
        # used to ensure that only a single persistent operation happens
        self.m_persistence_save_lock = threading.Lock()
        
        # condition variable and termination flag for blocking on a RPC result
        self.m_terminating = False
        self.m_rpc_results_cond = threading.Condition()
        
        # configured by dispatcher at startup
        self.m_queue_outgoing = None
        self.m_server = True
        
        self._process_fns = {
            kEntryAssign:   self._processEntryAssign,
            kEntryUpdate:   self._processEntryUpdate,
            kFlagsUpdate:   self._processFlagsUpdate,
            kEntryDelete:   self._processEntryDelete,
            kClearEntries:  self._processClearEntries,
            kExecuteRpc:    self._processExecuteRpc,
            kRpcResponse:   self._processRpcResponse
        }
    
    def stop(self):
        self.m_terminating = True
        with self.m_rpc_results_cond:
            self.m_rpc_results_cond.notify_all()
    
    def setOutgoing(self, queue_outgoing, server):
        with self.m_mutex:
            self.m_queue_outgoing = queue_outgoing
            self.m_server = server
    
    def clearOutgoing(self):
        self.m_queue_outgoing = None
    
    def getEntryType(self, msg_id):
        with self.m_mutex:
            if msg_id >= len(self.m_idmap):
                return NT_UNASSIGNED
        
            entry = self.m_idmap[msg_id]
            if not entry or not entry.value:
                return NT_UNASSIGNED
        
            return entry.value.type
    
    
    def processIncoming(self, msg, conn):
        # Note: c++ version takes a third param (weak_conn), but it's
        #       not needed here as conn == weak_conn
        fn = self._process_fns.get(msg.type)
        if fn:
            
            with self.m_mutex:
                queue_outgoing = self.m_queue_outgoing
                outgoing = [] if queue_outgoing else None
                fn(msg, conn, outgoing)
                    
            # this has to happen outside the lock
            if outgoing:
                for o in outgoing:
                    queue_outgoing(*o)
    
    def _processEntryAssign(self, msg, conn, outgoing):

        msg_id = msg.id
        name = msg.str
        entry = None
        may_need_update = False
        
        if self.m_server:
            # if we're a server, id=0xffff requests are requests for an id
            # to be assigned, we need to send the assignment back to
            # the sender as well as all other connections.
            if msg_id == 0xffff:
                # see if it was already assigned; ignore if so.
                entry = self.m_entries.get(name)
                if name in self.m_entries:
                    return
                
                # create it locally
                entry = _Entry(name, msg.value)
                msg_id = len(self.m_idmap)
                self.m_entries[name] = entry
                
                entry.flags = msg.flags
                entry.id = msg_id
                self.m_idmap.append(entry)

                # update persistent dirty flag if it's persistent
                if entry.isPersistent():
                    self.m_persistent_dirty = True
                
                # notify
                self.m_notifier.notifyEntry(name, entry.value, NT_NOTIFY_NEW)

                # send the assignment to everyone (including the originator)
                if outgoing is not None:
                    outmsg = Message.entryAssign(name, msg_id, entry.seq_num, msg.value, msg.flags)
                    outgoing.append((outmsg, None, None))

                return

            if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
                # ignore arbitrary entry assignments
                # this can happen due to e.g. assignment to deleted entry
                logger.debug("server: received assignment to unknown entry")
                return

            entry = self.m_idmap[msg_id]

        else:
            # clients simply accept assignments
            if msg_id == 0xffff:
                logger.debug("client: received entry assignment request?")
                return

            ensure_id_exists(self.m_idmap, msg_id)
            
            entry = self.m_idmap[msg_id]
            if not entry:
                # create local
                entry = self.m_entries.get(name)
                if not entry:
                    # didn't exist at all (rather than just being a response to a
                    # id assignment request)
                    entry = _Entry(name, msg.value)
                    entry.flags = msg.flags
                    entry.id = msg_id
                    self.m_entries[name] = entry
                    self.m_idmap[msg_id] = entry

                    # notify
                    self.m_notifier.notifyEntry(name, entry.value, NT_NOTIFY_NEW)
                    return

                may_need_update = True;  # we may need to send an update message
                entry.id = msg_id
                self.m_idmap[msg_id] = entry

                # if the received flags don't match what we sent, most likely
                # updated flags locally in the interim; send flags update message.
                if msg.flags != entry.flags and outgoing is not None:
                    outmsg = Message.flagsUpdate(msg_id, entry.flags)
                    outgoing.append((outmsg, None, None))


        # common client and server handling

        # already exists; ignore if sequence number not higher than local
        seq_num = msg.seq_num_uid
        if entry.isSeqNewerThan(seq_num):
            if may_need_update and outgoing is not None:
                outmsg = Message.entryUpdate(entry.id, entry.seq_num,
                                             entry.value)
                outgoing.append((outmsg, None, None))
            
            return

        # sanity check: name should match id
        if msg.str != entry.name:
            logger.debug("entry assignment for same id with different name?")
            return

        notify_flags = NT_NOTIFY_UPDATE

        # don't update flags from a <3.0 remote (not part of message)
        # don't update flags if self is a server response to a client id request
        if not may_need_update and conn.get_proto_rev() >= 0x0300:
            # update persistent dirty flag if persistent flag changed
            if (entry.flags & NT_PERSISTENT) != (msg.flags & NT_PERSISTENT):
                self.m_persistent_dirty = True

            if entry.flags != msg.flags:
                notify_flags |= NT_NOTIFY_FLAGS

            entry.flags = msg.flags


        # update persistent dirty flag if the value changed and it's persistent
        if entry.isPersistent() and entry.value != msg.value:
            self.m_persistent_dirty = True


        # update local
        entry.value = msg.value
        entry.seq_num = seq_num

        # notify
        self.m_notifier.notifyEntry(name, entry.value, notify_flags)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outmsg = Message.entryAssign(entry.name, msg_id, msg.seq_num_uid,
                                         msg.value, entry.flags)
            outgoing.append((outmsg, None, conn))
    
    def _processEntryUpdate(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates
            # self can happen due to deleted entries
            logger.debug("received update to unknown entry")
            return

        entry = self.m_idmap[msg_id]

        # ignore if sequence number not higher than local
        seq_num = msg.seq_num_uid
        if entry.isSeqNewerThan(seq_num):
            return
        
        # update local
        entry.value = msg.value
        entry.seq_num = seq_num

        # update persistent dirty flag if it's a persistent value
        if entry.isPersistent():
            self.m_persistent_dirty = True
        
        # notify
        self.m_notifier.notifyEntry(entry.name, entry.value, NT_NOTIFY_UPDATE)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processFlagsUpdate(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates
            # self can happen due to deleted entries
            logger.debug("received flags update to unknown entry")
            return

        entry = self.m_idmap[msg_id]

        # ignore if flags didn't actually change
        if entry.flags == msg.flags:
            return


        # update persistent dirty flag if persistent flag changed
        if (entry.flags & NT_PERSISTENT) != (msg.flags & NT_PERSISTENT):
            self.m_persistent_dirty = True


        # update local
        entry.flags = msg.flags

        # notify
        self.m_notifier.notifyEntry(entry.name, entry.value, NT_NOTIFY_FLAGS)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processEntryDelete(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates
            # self can happen due to deleted entries
            logger.debug("received delete to unknown entry")
            return

        entry = self.m_idmap[msg_id]

        # update persistent dirty flag if it's a persistent value
        if entry.isPersistent():
            self.m_persistent_dirty = True


        # delete it from idmap
        self.m_idmap[msg_id] = None

        # get entry (as we'll need it for notify) and erase it from the map
        try:
            entry2 = self.m_entries.pop(entry.name)
        except KeyError:
            # it should always be in the map, sanity check just in case
            pass
        else:
            self.m_notifier.notifyEntry(entry2.name, entry2.value, NT_NOTIFY_DELETE)
        
        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processClearEntries(self, msg, conn, outgoing):
        # update local
        self._deleteAllEntriesImpl()

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processExecuteRpc(self, msg, conn, outgoing):
        if not self.m_server:
            return    # only process on server

        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore call to non-existent RPC
            # self can happen due to deleted entries
            logger.debug("received RPC call to unknown entry")
            return

        entry = self.m_idmap[msg_id]
        if not entry.value.IsRpc():
            logger.debug("received RPC call to non-RPC entry")
            return

        self.m_rpc_server.processRpc(entry.name, msg, entry.rpc_callback,
                                     conn.uid(), conn.queueOutgoing, conn.info())
    
    def _processRpcResponse(self, msg, conn, outgoing):
        if self.m_server:
            return    # only process on client
        
        self.m_rpc_results[(msg.id, msg.seq_num_uid)] = msg.str
        self.m_rpc_results_cond.notify_all()
    
    def getInitialAssignments(self, conn, msgs):
        with self.m_mutex:
            conn.set_state(NetworkConnection.State.kSynchronized)
            for entry in self.m_entries.values():
                msgs.append(Message.entryAssign(entry.name, entry.id,
                                                entry.seq_num,
                                                entry.value, entry.flags))
    
    def applyInitialAssignments(self, conn, msgs, new_server, out_msgs):
        with self.m_mutex:
            if self.m_server:
                return    # should not do this on server
        
            conn.set_state(NetworkConnection.State.kSynchronized)
        
            update_msgs = []
        
            # clear existing id's
            for entry in self.m_entries.values():
                entry.id = 0xffff
        
            # clear existing idmap
            del self.m_idmap[:]
        
            # apply assignments
            for msg in msgs:
                if msg.type != kEntryAssign:
                    logger.debug("client: received non-entry assignment request?")
                    continue
        
                msg_id = msg.id
                if msg_id == 0xffff:
                    logger.debug("client: received entry assignment request?")
                    continue
        
                seq_num = msg.seq_num_uid
                name = msg.str
        
                entry = self.m_entries.get(name)
                if not entry:
                    # doesn't currently exist
                    entry = _Entry(name, msg.value, msg.flags, seq_num)
                    self.m_entries[name] = entry
                    
                    # notify
                    self.m_notifier.notifyEntry(name, entry.value, NT_NOTIFY_NEW)
        
                else:
                    # if reconnect and sequence number not higher than local, we
                    # don't update the local value and instead send it back to the server
                    # as an update message
                    if not new_server and entry.isSeqNewerThan(seq_num):
                        update_msgs.append(Message.entryUpdate(entry.id, entry.seq_num, entry.value))
        
                    else:
                        entry.value = msg.value
                        entry.seq_num = seq_num
                        notify_flags = NT_NOTIFY_UPDATE
                        # don't update flags from a <3.0 remote (not part of message)
                        if conn.get_proto_rev() >= 0x0300:
                            if entry.flags != msg.flags:
                                notify_flags |= NT_NOTIFY_FLAGS
        
                            entry.flags = msg.flags
        
                        # notify
                        self.m_notifier.notifyEntry(name, entry.value, notify_flags)
        
                # set id and save to idmap
                entry.id = msg_id
                
                ensure_id_exists(self.m_idmap, msg_id)
                self.m_idmap[msg_id] = entry
            
            # generate assign messages for unassigned local entries
            for entry in self.m_entries.values():
                if entry.id != 0xffff:
                    continue
        
                out_msgs.append(Message.entryAssign(entry.name, entry.id,
                                                    entry.seq_num,
                                                    entry.value, entry.flags))
        
            queue_outgoing = self.m_queue_outgoing
        
        # Outside of mutex
        for msg in update_msgs:
            queue_outgoing(msg, None, None)
    
    def getEntryValue(self, name):
        with self.m_mutex:
            e = self.m_entries.get(name)
            if e:
                return e.value
    
    def setDefaultEntryValue(self, name, value):
        if not value:
            return False    # can't compare to a null value
    
        if not name:
            return False    # can't compare empty name
    
        with self.m_mutex:
            entry = self.m_entries.get(name)
            if entry:   # entry already exists
                old_value = entry.value
                
                # if types match return True
                if old_value.type == value.type:
                    return True
                else:
                    return False    # entry exists but doesn't match type
        
            # if we've gotten here, does not exist, we can write it.
            # don't need to compare old value as we know it will assign
            entry = _Entry(name, value)
            self.m_entries[name] = entry
            
            # if we're the server, an id if it doesn't have one
            if self.m_server and entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
        
            # notify (for local listeners)
            if self.m_notifier.m_local_notifiers:
                # always a new entry if we got this far
                self.m_notifier.notifyEntry(name, value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
        
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return True
            
            msg = Message.entryAssign(name, entry.id, entry.seq_num,
                                            entry.value, entry.flags)
            
        # Outside of mutex
        queue_outgoing(msg, None, None)
        return True
    
    def setEntryValue(self, name, value):
        if not name:
            return True
    
        if not value:
            return True
    
        with self.m_mutex:
            entry = self.m_entries.get(name)
            if not entry:
                old_value = None
                entry = _Entry(name, value)
                self.m_entries[name] = entry
            else:
                old_value = entry.value
                if old_value.type != value.type:
                    return False    # error on type mismatch
                
                entry.value = value
            
            # if we're the server, an id if it doesn't have one
            if self.m_server and entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
            
            # update persistent dirty flag if value changed and it's persistent
            if entry.isPersistent() and old_value != value:
                self.m_persistent_dirty = True
            
            # notify (for local listeners)
            if self.m_notifier.m_local_notifiers:
                if old_value is None:
                    self.m_notifier.notifyEntry(name, value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
        
                elif old_value != value:
                    self.m_notifier.notifyEntry(name, value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)
            
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return True
        
            msg = None
            
            if old_value is None:
                msg = Message.entryAssign(name, entry.id, entry.seq_num,
                                                value, entry.flags)
        
            elif old_value != value:
                entry.increment_seqnum()
                
                # don't send an update if we don't have an assigned id yet
                if entry.id != 0xffff:
                    msg = Message.entryUpdate(entry.id, entry.seq_num, value)
        
        # unlocked mutex
        if msg:
            queue_outgoing(msg, None, None)
            
        return True
    
    def setEntryTypeValue(self, name, value):
        if not name:
            return
    
        if not value:
            return
    
        with self.m_mutex:
            entry = self.m_entries.get(name)
            if not entry:
                old_value = None
                entry = _Entry(name, value)
                self.m_entries[name] = entry
            else:
                old_value = entry.value
                if old_value == value:
                    return
                
                entry.value = value
            
            # if we're the server, an id if it doesn't have one
            if self.m_server and entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
            
            # update persistent dirty flag if it's a persistent value
            if entry.isPersistent():
                self.m_persistent_dirty = True
            
            # notify (for local listeners)
            if self.m_notifier.m_local_notifiers:
                if old_value is None:
                    self.m_notifier.notifyEntry(name, value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
        
                else:
                    self.m_notifier.notifyEntry(name, value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)
            
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return
            
            msg = None
            
            if old_value is None or old_value.type != value.type:
                entry.increment_seqnum()
                
                msg = Message.entryAssign(name, entry.id, entry.seq_num,
                                                value, entry.flags)
            else:
                entry.increment_seqnum()
                
                # don't send an update if we don't have an assigned id yet
                if entry.id != 0xffff:
                    msg = Message.entryUpdate(entry.id, entry.seq_num, value)
            
        # unlocked mutex
        if msg:
            queue_outgoing(msg, None, None)
    
    def setEntryFlags(self, name, flags):
        if not name:
            return
    
        with self.m_mutex:
            entry = self.m_entries.get(name)
            if not entry:
                return
        
            if entry.flags == flags:
                return
            
            # update persistent dirty flag if persistent flag changed
            if (entry.flags & NT_PERSISTENT) != (flags & NT_PERSISTENT):
                self.m_persistent_dirty = True
            
            entry.flags = flags
        
            # notify
            self.m_notifier.notifyEntry(name, entry.value, NT_NOTIFY_FLAGS | NT_NOTIFY_LOCAL)
        
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return
            
            entry_id = entry.id
            
        # don't send an update if we don't have an assigned id yet
        if entry_id != 0xffff:
            queue_outgoing(Message.flagsUpdate(entry_id, flags), None, None)
    
    def getEntryFlags(self, name):
        with self.m_mutex:
            entry = self.m_entries.get(name)
            return entry.flags if entry else 0
    
    def deleteEntry(self, name):
        with self.m_mutex:
            entry = self.m_entries.pop(name, None)
            if not entry:
                return
        
            entry_id = entry.id
        
            # update persistent dirty flag if it's a persistent value
            if entry.isPersistent():
                self.m_persistent_dirty = True
            
            if entry_id < len(self.m_idmap):
                self.m_idmap[entry_id] = None
            
            if not entry.value:
                return
            
            # notify
            self.m_notifier.notifyEntry(name, entry.value,
                                        NT_NOTIFY_DELETE | NT_NOTIFY_LOCAL)
        
            # if it had a value, message
            # don't send an update if we don't have an assigned id yet
            queue_outgoing = self.m_queue_outgoing
            
        if entry_id != 0xffff and queue_outgoing:
            queue_outgoing(Message.entryDelete(entry_id), None, None)
    
    def _deleteAllEntriesImpl(self):
        if not self.m_entries:
            return
    
        # only delete non-persistent values
        # can't erase without invalidating iterators, grab a list of keys
        for k in list(self.m_entries.keys()):
            entry = self.m_entries.get(k)
            if not entry.isPersistent():
                # notify it's being deleted
                if self.m_notifier.m_local_notifiers:
                    self.m_notifier.notifyEntry(k, entry,
                                                NT_NOTIFY_DELETE | NT_NOTIFY_LOCAL)
    
                # remove it from idmap
                if entry.id != 0xffff:
                    self.m_idmap[entry.id] = None
            
                # Delete it
                self.m_entries.pop(k)
    
    def deleteAllEntries(self):
        with self.m_mutex:
            if not self.m_entries:
                return
        
            self._deleteAllEntriesImpl()
        
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return
        
        queue_outgoing(Message.clearEntries(), None, None)
    
    def getEntryInfo(self, prefix, types):
        with self.m_mutex:
            infos = []
            types = types if isinstance(types, int) else ord(types)
            for k, entry in self.m_entries.items():
                if not k.startswith(prefix):
                    continue
                
                value = entry.value
                if not value:
                    continue
        
                if types != 0 and (types & ord(value.type)) == 0:
                    continue
        
                info = EntryInfo(entry.name, value.type, entry.flags)
                infos.append(info)
        
            return infos
    
    def notifyEntries(self, prefix, only):
        with self.m_mutex:
            
            for k, entry in self.m_entries.items():
                if not k.startswith(prefix):
                    continue
                
                self.m_notifier.notifyEntry(k, entry.value, NT_NOTIFY_IMMEDIATE,
                                            only)
    
    def _putPersistentEntries(self, entries):
        
        msgs = []
        
        # copy values into storage as quickly as possible so lock isn't held
        with self.m_mutex:
            queue_outgoing = self.m_queue_outgoing
            for name, value in entries:
                entry = self.m_entries.get(name)
                if not entry:
                    entry = _Entry(name)
                    self.m_entries[name] = entry
    
                old_value = entry.value
                entry.value = value
                was_persist = entry.isPersistent()
                if not was_persist:
                    entry.flags |= NT_PERSISTENT
                
                # if we're the server, an id if it doesn't have one
                if self.m_server and entry.id == 0xffff:
                    entry.id = len(self.m_idmap)
                    self.m_idmap.append(entry)
                
                # notify (for local listeners)
                if self.m_notifier.m_local_notifiers:
                    if old_value is None:
                        self.m_notifier.notifyEntry(name, value,
                                                    NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
                    elif old_value != value:
                        notify_flags = NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL
                        if not was_persist:
                            notify_flags |= NT_NOTIFY_FLAGS
    
                        self.m_notifier.notifyEntry(name, value, notify_flags)
    
                if not queue_outgoing:
                    continue    # shortcut
    
                entry.increment_seqnum()
    
                # put on update queue
                if old_value is None or old_value.type != value.type:
                    msgs.append(Message.entryAssign(name, entry.id,
                                                    entry.seq_num,
                                                    value, entry.flags))
                elif entry.id != 0xffff:
                    # don't send an update if we don't have an assigned id yet
                    if old_value != value:
                        msgs.append(Message.entryUpdate(entry.id, entry.seq_num, value))
                    if not was_persist:
                        msgs.append(Message.flagsUpdate(entry.id, entry.flags))
            
        if queue_outgoing:
            for msg in msgs:
                queue_outgoing(msg, None, None)
    
    def _getPersistentEntries(self, periodic):
        
        # copy values out of storage as quickly as possible so lock isn't held
        with self.m_mutex:
            if periodic and not self.m_persistent_dirty:
                return False
            
            self.m_persistent_dirty = False
            
            entries = [(entry.name, entry.value) for entry in self.m_entries.values() \
                       if entry.isPersistent()]
        
        # sort in name order
        entries.sort()
        return entries
    
    def loadPersistent(self, filename=None, fp=None):
        try:
            if fp:
                entries = load_entries(fp, filename if filename else '<string>')
            else:
                with open(filename, 'r') as fp:
                    entries = load_entries(fp, filename)
        except IOError as e:
            return 'Error reading file: %s' % e
        else:
            self._putPersistentEntries(entries)
            return
    
    def savePersistent(self, filename=None, periodic=False, fp=None):
        with self.m_persistence_save_lock:
            entries = self._getPersistentEntries(periodic)
            if entries is False:
                return
        
            # Going to not use tempfile to keep compatibility with ntcore,
            # as having to cleanup temp files on the RIO is probably bad
            if fp:
                save_entries(fp, entries)
            else:
                tmp = '%s.tmp' % filename
                bak = '%s.bak' % filename
                
                try:
                    with open(tmp, 'w') as fp:
                        save_entries(fp, entries)
                        os.fsync(fp.fileno())
                except IOError as e:
                    return 'Error writing file: %s' % e
                
                try:
                    file_replace(filename, bak)
                except OSError:
                    pass # ignored
                    
                try:
                    file_replace(tmp, filename)
                except OSError as e:
                    try:
                        # try to restore backup
                        file_replace(bak, filename)
                    except OSError:
                        pass
                        
                    return 'Could not rename temp file to real file: %s' % e
            
            self.m_persistent_dirty = False
    
    def createRpc(self, name, defn, callback):
        if not name or defn or not callback:
            return
    
        with self.m_mutex:
            if not self.m_server:
                return    # only server can create RPCs
            
            entry = self.m_entries.get(name)
            if not entry:
                entry = _Entry(name)
                self.m_entries[name] = entry
            
            old_value = entry.value
            value = Value.makeRpc(defn)
            entry.value = value
        
            # set up the callback
            entry.rpc_callback = callback
        
            # start the RPC server
            self.m_rpc_server.start()
        
            if old_value == value:
                return
            
            # assign an id if it doesn't have one
            if entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
            
            # generate message
            if not self.m_queue_outgoing:
                return
        
            queue_outgoing = self.m_queue_outgoing
            msg = None
            
            if old_value is None or old_value.type != value.type:
                entry.increment_seqnum()
                msg = Message.entryAssign(name, entry.id, entry.seq_num,
                                                value, entry.flags)
            else:
                entry.increment_seqnum()
                msg = Message.entryUpdate(entry.id, entry.seq_num, value)
                
        # unlocked mutex
        if msg:
            queue_outgoing(msg, None, None)
    
    def createPolledRpc(self, name, defn):
        if not name or not defn:
            return
        
        with self.m_mutex:
            if not self.m_server:
                return    # only server can create RPCs
            
            entry = self.m_entries.get(name)
            if not entry:
                entry = _Entry(name)
                self.m_entries[name] = entry
            
            old_value = entry.value
            value = Value.makeRpc(defn)
            entry.value = value
        
            # a None callback indicates a polled RPC
            entry.rpc_callback = None
        
            if old_value == value:
                return
            
            # assign an id if it doesn't have one
            if entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
            
            # generate message
            queue_outgoing = self.m_queue_outgoing
            if not queue_outgoing:
                return
            
            msg = None
            
            if old_value is None or old_value.type != value.type:
                entry.increment_seqnum()
                msg = Message.entryAssign(name, entry.id, entry.seq_num,
                                                value, entry.flags)
            else:
                entry.increment_seqnum()
                msg = Message.entryUpdate(entry.id, entry.seq_num, value)
                
        # unlocked mutex
        if msg:
            queue_outgoing(msg, None, None)
    
    def callRpc(self, name, params):
        self.m_mutex.acquire()
        locked = True
        try:
            entry = self.m_entries.get(name)
            if not entry:
                return 0
            
            if not entry.value.isRpc():
                return 0
        
            entry.rpc_call_uid += 1
            entry.rpc_call_uid &= 0xffff
            
            combined_uid = (entry.id << 16) | entry.rpc_call_uid
            msg = Message.executeRpc(entry.id, entry.rpc_call_uid, params)
            if self.m_server:
                # RPCs are unlikely to be used locally on the server, handle it
                # gracefully anyway.
                rpc_callback = entry.rpc_callback
                
                self.m_mutex.release()
                locked = False
                
                conn_info = ConnectionInfo('Server', 'localhost', 0, monotonic(), 0x0300)
                self.m_rpc_server.processRpc(name, msg, rpc_callback, 0xffff,
                                             self._process_rpc, conn_info)
            else:
                queue_outgoing = self.m_queue_outgoing
                
                self.m_mutex.release()
                locked = False
                
                queue_outgoing(msg, None, None)
        
            return combined_uid
        finally:
            if locked:
                self.m_mutex.release()
    
    def _process_rpc(self, msg):
        with self.m_mutex:
            self.m_rpc_results[(msg.id, msg.seq_num_uid)] = msg.str
            self.m_rpc_results_cond.notify_all()
    
    def getRpcResult(self, blocking, call_uid, time_out=-1):
        with self.m_mutex:
            # only allow one blocking call per rpc call uid
            if call_uid in self.m_rpc_blocking_calls:
                return False, None
        
            self.m_rpc_blocking_calls.add(call_uid)
            wait_until = monotonic() + time_out
        
            try:
                while True:
                    result = self.m_rpc_results.get(call_uid) 
                    if not result:
                        if not blocking or self.m_terminating:
                            return False, None
            
                        if time_out <= 0:
                            self.m_rpc_results_cond.wait()
                        else:
                            ttw = monotonic() - wait_until
                            if ttw <= 0:
                                return False, None
                            
                            self.m_rpc_results_cond.wait(ttw)
            
                        # if element does not exist, have been canceled
                        if call_uid not in self.m_rpc_blocking_calls:
                            return False, None
            
                        if self.m_terminating:
                            return False, None
            
                        continue
                    
                    self.m_rpc_results.pop(call_uid, None)
                    return True, result
            finally:
                try:
                    self.m_rpc_blocking_calls.remove(call_uid)
                except KeyError:
                    pass
    
    def cancelBlockingRpcResult(self, call_uid):
        with self.m_mutex:
            # safe to erase even if id does not exist
            self.m_rpc_blocking_calls.erase(call_uid)
            self.m_rpc_results_cond.notify_all()
    
    
    
   



# novalidate

import threading
import socket

from .tcp_stream import TCPStream

import logging
logger = logging.getLogger('nt')

class TcpAcceptor(object):
    
    def __init__(self, port, address):
        # Protects open/shutdown/close
        # -> This is a condition to allow testing code to wait
        #    for server startup
        self.lock = threading.Condition()
        
        self.m_lsd = None
        self.m_port = port
        self.m_address = address
        self.m_listening = False
        self.m_shutdown = False
        
    def waitForStart(self, timeout=None):
        with self.lock:
            if not self.m_listening:
                self.lock.wait(timeout=timeout)
        
        return self.m_listening
        
    def close(self):
        
        with self.lock:
            if self.m_lsd:
                self.shutdown()
                self.m_lsd.close()
            
            self.m_lsd = None
    
    def start(self):
        with self.lock:
            if self.m_listening:
                return False
            
            self.m_lsd = socket.socket()
            
            try:
                self.m_lsd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                self.m_lsd.bind((self.m_address, self.m_port))
                
                # needed for testing
                if self.m_port == 0:
                    self.m_port = self.m_lsd.getsockname()[1]
                
                self.m_lsd.listen(10)
            except OSError:
                logger.exception("Error starting server")
                
                try:
                    self.m_lsd.close()
                except Exception:
                    pass
                
                self.m_lsd = None
                self.lock.notify()
                return False
            
            self.m_listening = True
            self.lock.notify()
            logger.debug("Listening on %s %s", self.m_address, self.m_port)
            return True
    
    def shutdown(self):
        with self.lock:
            if self.m_listening and not self.m_shutdown:
                self.m_shutdown = True
                self.m_listening = False
                try:
                    self.m_lsd.shutdown(socket.SHUT_RDWR)
                except OSError:
                    pass
    
    def accept(self):
        if not self.m_listening or self.m_shutdown:
            return
        
        try:
            sd, (peer_ip, peer_port) = self.m_lsd.accept()
        except OSError:
            if not self.m_shutdown:
                logger.warning("Error accepting connection", exc_info=True)
            return
        
        if self.m_shutdown:
            try:
                sd.close()
            except Exception:
                pass
            return
        
        return TCPStream(sd, peer_ip, peer_port)



# validated: 2016-10-21 DS a73166a src/Value.cpp
'''
    Internal storage for ntcore values
    
    Uses namedtuple for efficiency, and because Value objects are supposed
    to be immutable. Will have to measure that and see if there's a performance
    penalty for this in python.
    
    Original ntcore stores the last change time, but it doesn't seem to
    be used anywhere, so we don't store that to make equality comparison
    more efficient.
'''

from collections import namedtuple
from .constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    NT_RPC,
)

from .support.compat import basestring, unicode

ValueType = namedtuple('Value', ['type', 'value'])


# optimization
_TRUE_VALUE = ValueType(NT_BOOLEAN, True)
_FALSE_VALUE = ValueType(NT_BOOLEAN, False)


class Value(object):
    
    @staticmethod
    def makeBoolean(value):
        if value:
            return _TRUE_VALUE
        else:
            return _FALSE_VALUE
        
    @staticmethod
    def makeDouble(value):
        return ValueType(NT_DOUBLE, float(value))

    @staticmethod
    def makeString(value):
        return ValueType(NT_STRING, unicode(value))
    
    @staticmethod
    def makeRaw(value):
        return ValueType(NT_RAW, bytes(value))
    
    # TODO: array stuff a good idea?
    
    @staticmethod
    def makeBooleanArray(value):
        return ValueType(NT_BOOLEAN_ARRAY, tuple(bool(v) for v in value))
    
    @staticmethod
    def makeDoubleArray(value):
        return ValueType(NT_DOUBLE_ARRAY, tuple(float(v) for v in value))
    
    @staticmethod
    def makeStringArray(value):
        return ValueType(NT_STRING_ARRAY, tuple(unicode(v) for v in value))

    @staticmethod
    def makeRpc(value):
        return ValueType(NT_RPC, str(value))

    @staticmethod
    def getFactory(value):
        if isinstance(value, bool):
            return Value.makeBoolean
        elif isinstance(value, (int, float)):
            return Value.makeDouble
        elif isinstance(value, basestring):
            return Value.makeString
        elif isinstance(value, bytes):
            return Value.makeRaw
        
        # Do best effort for arrays, but can't catch all cases
        # .. if you run into an error here, use a less generic type
        elif isinstance(value, (list, tuple)):
            if not value:
                raise TypeError("If you use a list here, cannot be empty")
            
            first = value[0]
            if isinstance(first, bool):
                return Value.makeBooleanArray
            elif isinstance(first, (int, float)):
                return Value.makeDoubleArray
            elif isinstance(first, basestring):
                return Value.makeStringArray
            else:
                raise ValueError("Can only use lists of bool/int/float/strings")
        
        elif value is None:
            raise ValueError("Cannot put None into NetworkTable")
        else:
            raise ValueError("Can only put bool/int/float/str/bytes or lists/tuples of them")



# validated: 2016-10-21 DS c0ce427 src/ntcore_cpp.cpp

from .dispatcher import Dispatcher
from .notifier import Notifier
from .rpc_server import RpcServer
from .storage import Storage

from .constants import NT_NOTIFY_IMMEDIATE

class NtCoreApi(object):
    '''
        Internal NetworkTables API wrapper
        
        In theory you could create multiple instances of this
        and talk to multiple NT servers or create multiple
        NT servers... though, I don't really know why one
        would want to do this.
    '''
    
    def __init__(self, verbose=False):
        self.notifier = Notifier(verbose=verbose)
        self.rpc_server = RpcServer()
        self.storage = Storage(self.notifier, self.rpc_server)
        self.dispatcher = Dispatcher(self.storage, self.notifier, verbose=verbose)
        
    def stop(self):
        self.dispatcher.stop()
        self.rpc_server.stop()
        self.notifier.stop()
        self.storage.stop()

    #
    # Table functions
    #
    
    def getEntryValue(self, name):
        return self.storage.getEntryValue(name)
    
    def setDefaultEntryValue(self, name, value):
        return self.storage.setDefaultEntryValue(name, value)
    
    def setEntryValue(self, name, value):
        return self.storage.setEntryValue(name, value)
    
    def setEntryTypeValue(self, name, value):
        self.storage.setEntryTypeValue(name, value)
    
    def setEntryFlags(self, name, flags):
        self.storage.setEntryFlags(name, flags)
    
    def getEntryFlags(self, name):
        return self.storage.getEntryFlags(name)
    
    def deleteEntry(self, name):
        self.storage.deleteEntry(name)
    
    def deleteAllEntries(self):
        self.storage.deleteAllEntries()
    
    def getEntryInfo(self, prefix, types):
        return self.storage.getEntryInfo(prefix, types)
    
    def flush(self):
        self.dispatcher.flush()
    
    #
    # Callback creation functions
    #
    
    def setListenerOnStart(self, on_start):
        self.notifier.setOnStart(on_start)
    
    def setListenerOnExit(self, on_exit):
        self.notifier.setOnExit(on_exit)
    
    def addEntryListener(self, prefix, callback, flags):
        uid = self.notifier.addEntryListener(prefix, callback, flags)
        if (flags & NT_NOTIFY_IMMEDIATE) != 0:
            self.storage.notifyEntries(prefix, callback)
    
        return uid
    
    def removeEntryListener(self, entry_listener_uid):
        self.notifier.removeEntryListener(entry_listener_uid)
    
    def addConnectionListener(self, callback, immediate_notify):
        self.notifier.addConnectionListener(callback)
        if immediate_notify:
            self.dispatcher.notifyConnections(callback)
    
    def removeConnectionListener(self, callback):
        self.notifier.removeConnectionListener(callback)

    # Python-specific
    def createAutoValue(self, key, value):
        return self.notifier.createAutoValue(key, value)
    
    #
    # TODO: RPC stuff not currently implemented
    #       .. there's probably a good pythonic way to implement
    #          it, but I don't really want to deal with it now.
    #          If you care, submit a PR.
    #
    #          I would have the caller register the server function
    #          via a docstring.
    #
    
    #
    # Initialization functions
    #
        
    def setNetworkIdentity(self, name):
        self.dispatcher.setIdentity(name)
        
    def startServer(self, persist_filename, listen_address, port):
        self.dispatcher.startServer(persist_filename, listen_address, port)
    
    def stopServer(self):
        self.dispatcher.stop()
        
    def startClient(self, servers):
        self.dispatcher.startClient(servers)
        
    def stopClient(self):
        self.dispatcher.stop()
    
    def setUpdateRate(self, interval):
        self.dispatcher.setUpdateRate(interval)
    
    def getRemoteAddress(self):
        if not self.dispatcher.isServer():
            for conn in self.dispatcher.getConnections():
                return conn.remote_ip
    
    def getIsConnected(self):
        return self.dispatcher.isConnected()
    
    def setVerboseLogging(self, verbose):
        self.dispatcher.setVerboseLogging(verbose)
        self.notifier.setVerboseLogging(verbose)
    
    #
    # Persistence
    #
    
    def savePersistent(self, filename):
        return self.storage.savePersistent(filename, periodic=False)
        
    def loadPersistent(self, filename):
        return self.storage.loadPersistent(filename)
    
    
        


#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# When running, this will continue incrementing the value 'dsTime', and the
# value should be visible to other networktables clients and the robot.
#

import sys
import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTables.initialize(server=ip)

sd = NetworkTables.getTable("SmartDashboard")

i = 0
while True:
    try:
        print('robotTime:', sd.getNumber('robotTime'))
    except KeyError:
        print('robotTime: N/A')

    sd.putNumber('dsTime', i)
    time.sleep(1)
    i += 1




# notrack
'''
    This implements an undocumented protocol that the LabVIEW dashboard
    uses to receive information from the driver station. In particular,
    it sends data about what IP the robot is at.
    
    To use this, call NetworkTable.setDashboardMode()
'''

import socket
import struct
import threading

try:
    import SocketServer as socketserver
except ImportError:
    import socketserver

from .socketstream import SocketStreamFactory

import logging
logger = logging.getLogger('nt.dashboard')

_tagdata = struct.Struct('!hb')
_u32 = struct.Struct('!i')


class DsDataHandler(socketserver.StreamRequestHandler):
    '''
        Handler that receives data from the driver station
    
        Data is received as a stream of the following:
        
        * 2 bytes big endian: length to follow
        * 1 byte: tag
        * n-1 bytes: content
        
        Tag number 8 is the roborio IP address, and there should
        be 4 bytes received, transmitted in network order.
    '''
    
    def setup(self):
        socketserver.StreamRequestHandler.setup(self)
        logger.info("Driver Station connected! Waiting for robot...")
    
    def handle(self):
        while True:
            data = self.rfile.read(3)
            if not data:
                break
            
            data_len, tag = _tagdata.unpack(data)
            
            data = self.rfile.read(data_len-1)
            if not data:
                break
            
            if tag == 8:
                address = socket.inet_ntoa(data)
                self.server.notifyConnection(address)
    
    def finish(self):
        socketserver.StreamRequestHandler.finish(self)
        logger.info("Driver Station disconnected!")

class DsDataServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    
    daemon_threads = True
    allow_reuse_address = True
    
    def __init__(self, f):
        socketserver.TCPServer.__init__(self, ('0.0.0.0', 1741), DsDataHandler)
        self.notifyConnection = f.notifyConnection
        
class DashboardSocketStreamFactory(SocketStreamFactory):
    
    def __init__(self, ipAddress, port):
        self.server = DsDataServer(self)
        
        self.host = None
        self.port = port
        
        self.lock = threading.Condition()
        self.thread = threading.Thread(target=self._run, name="DashboardListener")
        self.thread.daemon = True
        self.thread.start()
        
        logger.info("Dashboard mode enabled, listening for DS comms")
    
    def _run(self):
        self.server.serve_forever()
    
    def createStream(self):
        with self.lock:
            while self.host is None:
                self.lock.wait()
        
            return SocketStreamFactory.createStream(self)
    
    def notifyConnection(self, address):
        if address == '0.0.0.0':
            address = None
            self.host = None
            
        elif self.host != address:
            logger.info("Driver Station says robot IP is %s", address)
            
            with self.lock:
                self.host = address
                self.lock.notify()



#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# This shows how to use a listener to listen for all changes in NetworkTables
# values, which prints out all changes. Note that the keys are full paths, and
# not just individual key values.
#

import sys
import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

#if len(sys.argv) != 2:
    #print("Error: specify an IP to connect to!")
    #exit(0)

#ip = sys.argv[1]

#NetworkTables.initialize(server=ip)

def valueChanged(key, value, isNew):
    print("valueChanged: key: '%s'; value: %s; isNew: %s" % (key, value, isNew))

NetworkTables.addGlobalListener(valueChanged)

while True:
    time.sleep(1)

 



# validated: 2016-10-27 DS c2ae897 src/WireDecoder.cpp src/WireDecoder.h src/WireEncoder.cpp src/WireEncoder.h
'''
    This encompasses the WireEncoder and WireDecoder stuff in ntcore
    
    Reading:
    
    Writing:
    
        Each message type will have a write function, which takes
        a single list argument. Bytes will be added to that list.
    
        The write routines assume that the messages are a tuple
        that have the following format:
        
            # This doesn't make sense
            type, str, value, id, flags, seqnum
    
'''

import struct
from .support.compat import range

from .constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    NT_RPC
)

from .support import leb128
from .value import Value


_clientHello = struct.Struct('>H')
_protoUnsup = struct.Struct('>H')
_entryAssignV2 = struct.Struct('>HH')
_entryUpdate = struct.Struct('>HH')

_serverHello = struct.Struct('b')
_entryAssignV3 = struct.Struct('>HHb')
_flagsUpdate = struct.Struct('>Hb')
_entryDelete = struct.Struct('>H')
_clearEntries = struct.Struct('>I')
_executeRpc = struct.Struct('>HH')
_rpcResponse = struct.Struct('>HH')


class WireCodec(object):
    
    _bool_fmt = struct.Struct('?')
    _double_fmt = struct.Struct('>d')
    _string_fmt = struct.Struct('>H')
    _array_fmt = struct.Struct('b')
    _short_fmt = struct.Struct('>H')
        
    clientHello = _clientHello
    protoUnsup = _protoUnsup
    entryUpdate = _entryUpdate
        
    def __init__(self, proto_rev):
        self.set_proto_rev(proto_rev)
        
    def set_proto_rev(self, proto_rev):
        self.proto_rev = proto_rev
        if proto_rev == 0x0200:
            self.read_arraylen = self.read_arraylen_v2
            self.read_string = self.read_string_v2
            self.write_arraylen = self.write_arraylen_v2
            self.write_string = self.write_string_v2
            
            self.entryAssign = _entryAssignV2
            
            self._del('serverHello')
            self._del('flagsUpdate')
            self._del('entryDelete')
            self._del('clearEntries')
            self._del('executeRpc')
            self._del('rpcResponse')
            
        elif proto_rev == 0x0300:
            self.read_arraylen = self.read_arraylen_v3
            self.read_string = self.read_string_v3
            self.write_arraylen = self.write_arraylen_v3
            self.write_string = self.write_string_v3
            
            self.entryAssign = _entryAssignV3
            
            self.serverHello = _serverHello
            self.flagsUpdate = _flagsUpdate
            self.entryDelete = _entryDelete
            self.clearEntries = _clearEntries
            self.executeRpc = _executeRpc
            self.rpcResponse = _rpcResponse
        
        else:
            raise ValueError("Unsupported protocol")
    
    def _del(self, attr):
        if hasattr(self, attr):
            delattr(self, attr)
    
    def read_value(self, vtype, rstream):
        if vtype == NT_BOOLEAN:
            return Value.makeBoolean(rstream.readStruct(self._bool_fmt)[0])
        
        elif vtype == NT_DOUBLE:
            return Value.makeDouble(rstream.readStruct(self._double_fmt)[0])
        
        elif vtype == NT_STRING:
            return Value.makeString(self.read_string(rstream))
        
        elif vtype == NT_BOOLEAN_ARRAY:
            alen = self.read_arraylen(rstream)
            return Value.makeBooleanArray([rstream.readStruct(self._bool_fmt)[0] for _ in range(alen)])
        
        elif vtype == NT_DOUBLE_ARRAY:
            alen = self.read_arraylen(rstream)
            return Value.makeDoubleArray([rstream.readStruct(self._double_fmt)[0] for _ in range(alen)])
        
        elif vtype == NT_STRING_ARRAY:
            alen = self.read_arraylen(rstream)
            return Value.makeStringArray([self.read_string(rstream) for _ in range(alen)])
        
        elif self.proto_rev >= 0x0300:
            if vtype == NT_RAW:
                slen = leb128.read_uleb128(rstream) 
                return Value.makeRaw(rstream.read(slen))
            
            elif vtype == NT_RPC:
                return Value.makeRpc(self.read_string(rstream))
        
        raise ValueError("Cannot decode value type %s" % vtype)
    
    def write_value(self, v, out):
        vtype = v.type
    
        if vtype == NT_BOOLEAN:
            out.append(self._bool_fmt.pack(v.value))
            return
            
        elif vtype == NT_DOUBLE:
            out.append(self._double_fmt.pack(v.value))
            return
            
        elif vtype == NT_STRING:
            self.write_string(v.value, out)
            return
            
        elif vtype == NT_BOOLEAN_ARRAY:
            alen = self.write_arraylen(v.value, out)
            out += (self._bool_fmt.pack(v) for v in v.value[:alen])
            return
            
        elif vtype == NT_DOUBLE_ARRAY:
            alen = self.write_arraylen(v.value, out)
            out += (self._double_fmt.pack(v) for v in v.value[:alen])
            return
            
        elif vtype == NT_STRING_ARRAY:
            alen = self.write_arraylen(v.value, out)
            for s in v.value[:alen]:
                self.write_string(s, out)
            return
                
        elif self.proto_rev >= 0x0300:
            if vtype == NT_RPC:
                self.write_string(v.value, out)
                return
            
            elif vtype == NT_RAW:
                s = v.value
                out += (leb128.encode_uleb128(len(s)), s)
                return
        
        raise ValueError("Cannot encode invalid value type %s" % vtype)
    
    
    #
    # v2/v3 routines
    #
    
    def read_arraylen_v2(self, rstream):
        return rstream.readStruct(self._array_fmt)[0]
    
    def read_arraylen_v3(self, rstream):
        return leb128.read_uleb128(rstream) 
    
    def read_string_v2(self, rstream):
        slen = rstream.readStruct(self._string_fmt)[0]
        return rstream.read(slen).decode('utf-8', 'replace')
    
    def read_string_v3(self, rstream):
        slen = leb128.read_uleb128(rstream)
        return rstream.read(slen).decode('utf-8', 'replace')
    
    
    def write_arraylen_v2(self, a, out):
        alen = min(len(a), 0xff)
        out.append(self._array_fmt.pack(alen))
        return alen
    
    def write_arraylen_v3(self, a, out):
        alen = len(a)
        out.append(leb128.encode_uleb128(alen))
        return alen
    
    def write_string_v2(self, s, out):
        s = s.encode('utf-8')
        out += (self._string_fmt.pack(min(len(s), 0xffff)), s[:0xffff])
    
    def write_string_v3(self, s, out):
        s = s.encode('utf-8')
        out += (leb128.encode_uleb128(len(s)), s)





#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# When running, this will continue incrementing the value 'dsTime', and the
# value should be visible to other networktables clients and the robot.
#

import sys
import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTables.initialize(server=ip)

try:
    table = NetworkTables.getGlobalTable()
    
    i = 0
    while True:
        try:
            print('robotTime:', table.getNumber('/SmartDashboard/robotTime'))
        except KeyError:
            print('robotTime: N/A')
    
        table.putNumber('/SmartDashboard/dsTime', i)
        time.sleep(1)
        i += 1
    
finally:
    NetworkTables.shutdown()    



# notrack

import os
import sys

try:
    from configparser import RawConfigParser, NoSectionError
except ImportError:
    from ConfigParser import RawConfigParser, NoSectionError

try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

try:
    from time import monotonic
except ImportError:
    from monotonic import monotonic

if sys.version_info[0] <= 2:
    range = xrange
    basestring = basestring
    unicode = unicode
    PY2 = True
else:
    range = range
    basestring = str
    unicode = str
    PY2 = False
    
#
# For information about atomic writes, see
# -> http://stupidpythonideas.blogspot.com/2014/07/getting-atomic-writes-right.html
#
# Basically, if you're using Python 3.3+, good to go. Otherwise
# we'll try our best, but no guarantees.
#

if hasattr(os, 'replace'):      # Python 3.3+
    file_replace = os.replace
elif os.name != 'nt':           # Not Windows
    file_replace = os.rename
else:                           # Windows
    def file_replace(src, dst):
        try:
            os.unlink(dst)
        except FileNotFoundError:
            pass
        os.rename(src, dst)

#
# Polyfill for Condition.wait_for()
#
# Copied from Python 3.5 source code, Python license
#


if sys.version_info[0] >= 3: # technically, 3.2... but we don't support 3.2
    from threading import Condition
else:
    _time = monotonic
    from threading import _Condition as _ConditionBase
    
    class Condition(_ConditionBase):
        
        def wait_for(self, predicate, timeout=None):
            """Wait until a condition evaluates to True.
    
            predicate should be a callable which result will be interpreted as a
            boolean value.  A timeout may be provided giving the maximum time to
            wait.
    
            """
            endtime = None
            waittime = timeout
            result = predicate()
            while not result:
                if waittime is not None:
                    if endtime is None:
                        endtime = _time() + waittime
                    else:
                        waittime = endtime - _time()
                        if waittime <= 0:
                            break
                self.wait(waittime)
                result = predicate()
            return result



# novalidate


# data types
NT_UNASSIGNED =     b'\x00'
NT_BOOLEAN =        b'\x01'
NT_DOUBLE =         b'\x02'
NT_STRING =         b'\x04'
NT_RAW =            b'\x08'
NT_BOOLEAN_ARRAY =  b'\x10'
NT_DOUBLE_ARRAY =   b'\x20'
NT_STRING_ARRAY =   b'\x40'
NT_RPC =            b'\x80'

# Raw types transmitted on the wire
NT_VTYPE2RAW = {
    NT_BOOLEAN:       b'\x00',
    NT_DOUBLE:        b'\x01',
    NT_STRING:        b'\x02',
    NT_RAW:           b'\x03',
    NT_BOOLEAN_ARRAY: b'\x10',
    NT_DOUBLE_ARRAY:  b'\x11',
    NT_STRING_ARRAY:  b'\x12',
    NT_RPC:           b'\x20',
}

NT_RAW2VTYPE = {v: k for k, v in NT_VTYPE2RAW.items()}



# NetworkTables notifier kinds.
NT_NOTIFY_NONE =        0x00
NT_NOTIFY_IMMEDIATE =   0x01 # initial listener addition
NT_NOTIFY_LOCAL =       0x02 # changed locally
NT_NOTIFY_NEW =         0x04 # newly created entry
NT_NOTIFY_DELETE =      0x08 # deleted
NT_NOTIFY_UPDATE =      0x10 # value changed
NT_NOTIFY_FLAGS =       0x20 # flags changed

# NetworkTables entry flags
NT_PERSISTENT = 0x01


# Message types
kKeepAlive =        b'\x00'
kClientHello =      b'\x01'
kProtoUnsup =       b'\x02'
kServerHelloDone =  b'\x03'
kServerHello =      b'\x04'
kClientHelloDone =  b'\x05'
kEntryAssign =      b'\x10'
kEntryUpdate =      b'\x11'
kFlagsUpdate =      b'\x12'
kEntryDelete =      b'\x13'
kClearEntries =     b'\x14'
kExecuteRpc =       b'\x20'
kRpcResponse =      b'\x21'

kClearAllMagic =    0xD06CB27A




import threading

from ntcore import constants
from ntcore.api import NtCoreApi
from ntcore.value import Value

from .autovalue import AutoUpdateValue

import logging
logger = logging.getLogger('nt')

__all__ = ["NetworkTables"]


class NetworkTables:
    """
    This is the global singleton that you use to initialize NetworkTables
    connections, configure global settings and listeners, and to create
    NetworkTable instances which can be used to send data to/from
    NetworkTable servers and clients.
    
    First, you must initialize NetworkTables::
    
        from networktables import NetworkTables
        
        # As a client to connect to a robot
        NetworkTables.initialize(server='roborio-XXX-frc.local')
    

    Then, to interact with the SmartDashboard you get an instance of the
    table, and you can call the various methods::

        sd = NetworkTables.getTable('SmartDashboard')
        
        sd.putNumber('someNumber', 1234)
        otherNumber = sd.getNumber('otherNumber')
        ...
        
    .. seealso::
       - The examples in the documentation.
       - :class:`.NetworkTable`
    """
    
    class EntryTypes:
        '''
            NetworkTable value types used in :meth:`.NetworkTable.getKeys`
        '''
        
        #: True or False
        BOOLEAN = constants.NT_BOOLEAN
        
        #: Floating point number
        DOUBLE = constants.NT_DOUBLE
        
        #: Strings
        STRING = constants.NT_STRING
        
        #: Raw bytes
        RAW = constants.NT_RAW
        
        #: List of booleans
        BOOLEAN_ARRAY = constants.NT_BOOLEAN_ARRAY
        
        #: List of numbers
        DOUBLE_ARRAY = constants.NT_DOUBLE_ARRAY
        
        #: List of strings
        STRING_ARRAY = constants.NT_STRING_ARRAY
    
    class EntryFlags:
        '''
            NetworkTables entry flags
        '''
        
        #: Indicates a value that will be persisted on the server
        PERSISTENT = constants.NT_PERSISTENT
    
    class NotifyFlags:
        '''
            Bitflags passed to entry callbacks
        '''
        
        #: Initial listener addition
        IMMEDIATE = constants.NT_NOTIFY_IMMEDIATE
        
        #: Changed locally
        LOCAL = constants.NT_NOTIFY_LOCAL
        
        #: Newly created entry
        NEW = constants.NT_NOTIFY_NEW
        
        #: Key deleted
        DELETE = constants.NT_NOTIFY_DELETE
        
        #: Value changed
        UPDATE = constants.NT_NOTIFY_UPDATE
        
        #: Flags changed
        FLAGS = constants.NT_NOTIFY_FLAGS
    
    
    
    
    #: The path separator for sub-tables and keys
    PATH_SEPARATOR = '/'
    #: The default port that network tables operates on
    DEFAULT_PORT = 1735
    
    _persistentFilename = None
    
    port = DEFAULT_PORT
    ipAddress = None
    _serverListenAddress = ""
    
    _mode = 'server'
    _running = False
    _tables = {}

    _staticMutex = threading.RLock()

    _api = NtCoreApi()
    _identity = None

    @classmethod
    def _checkInit(cls):
        with cls._staticMutex:
            if cls._running:
                raise RuntimeError("Network tables has already been initialized")

    @classmethod
    def initialize(cls, server=None):
        """Initializes NetworkTables and begins operations
        
        :param server: If specified, NetworkTables will be set to client
                       mode and attempt to connect to the specified server.
                       This is equivalent to executing::
                       
                           cls.setIPAddress(server)
                           cls.setClientMode()
                           cls.initialize()
        :type server: str
        
        .. versionadded:: 2017.0.0
           The *server* parameter
        """
        
        with cls._staticMutex:
            cls._checkInit()
            
            # Circular import issue
            try:
                from .version import __version__
            except ImportError:
                __version__ = '[unknown version]'
            
            identity = cls._identity
            if identity is None:
                identity = 'pynetworktables %s' % __version__
            
            cls._api.setNetworkIdentity(identity)
            
            if server is not None:
                cls._mode = 'client'
                cls.ipAddress = server
            
            if cls._mode == 'server':
                cls._api.startServer(cls._persistentFilename,
                                     cls._serverListenAddress,
                                     cls.port)
            elif cls._mode == 'client':
                cls._api.startClient([(cls.ipAddress, cls.port)])
            elif cls._mode == 'dashboard':
                raise ValueError("Dashboard mode isn't implemented yet")
            elif cls._mode == 'test-server':
                pass
            elif cls._mode == 'test-client':
                cls._api.dispatcher.m_server = False
                cls._api.storage.m_server = False 
            else:
                raise ValueError("Invalid NetworkTables mode '%s'" % cls._mode)

            
            logger.info("NetworkTables %s initialized in %s mode",
                        __version__, cls._mode)

            cls._running = True
    
    @classmethod
    def shutdown(cls):
        """Stops all NetworkTables activities and unregisters all tables
        and callbacks. You can call :meth:`.initialize` again after
        calling this.
        
        .. versionadded:: 2017.0.0
        """
        
        with cls._staticMutex:
            
            try:
                if not cls._running:
                    return
                
                if cls._mode in ['client', 'dashboard']:
                    cls._api.stopClient()
                elif cls._mode == 'server':
                    cls._api.stopServer()
            finally:
                cls._running = False
                cls._persistentFilename = None
                cls._mode = 'server'
                cls._tables = {}
                cls.port = cls.DEFAULT_PORT
                cls.ipAddress = None
                cls._identity = None
                cls._autolistener = None
                cls._api = NtCoreApi()
    
    @classmethod
    def setClientMode(cls):
        """Set that network tables should be a client
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        with cls._staticMutex:
            cls._checkInit()
            cls._mode = 'client'
            
    @classmethod
    def setServerMode(cls):
        """set that network tables should be a server (this is the default)
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        with cls._staticMutex:
            cls._checkInit()
            cls._mode = 'server'
    
    @classmethod
    def setTeam(cls, team):
        """set the team the robot is configured for (this will set the ip
        address that network tables will connect to in client mode)
        
        :param team: the team number
        :type team: str
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        cls.setIPAddress('roboRIO-%d-FRC.local' % team)

    @classmethod
    def setIPAddress(cls, address):
        """:param address: the adress that network tables will connect to in
            client mode
        :type address: str
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        with cls._staticMutex:
            cls._checkInit()
            cls.ipAddress = address
    
    @classmethod
    def setPort(cls, port):
        """Sets the port number that network tables will connect to in
        client mode or listen to in server mode.
        
        :param port: the port number
        :type port: int
        
        .. versionadded:: 2017.0.0
        """
        with cls._staticMutex:
            cls._checkInit()
            cls.port = port
    
    @classmethod
    def setPersistentFilename(cls, filename):
        """Sets the persistent filename. Not used on the client.
         
        :param filename: the filename that the network tables server uses for
                         automatic loading and saving of persistent values
        :type filename: str
                         
        .. versionadded:: 2017.0.0
        """
        
        with cls._staticMutex:
            cls._checkInit()
            cls._persistentFilename = filename
        
    
    @classmethod
    def setNetworkIdentity(cls, name):
        """Sets the network identity. This is provided in the connection info
        on the remote end.
        
        :param name: A string to communicate to other NetworkTables instances
        :type name: str
        
        .. versionadded:: 2017.0.0
        """
        with cls._staticMutex:
            cls._checkInit()
            cls._identity = name
        
    @classmethod
    def globalDeleteAll(cls):
        """Deletes ALL keys in ALL subtables.
        
        .. warning:: Use with caution!
        
        .. versionadded:: 2017.0.0
        """
        cls._api.globalDeleteAll()
    
    @classmethod
    def flush(cls):
        """Flushes all updated values immediately to the network.
     
        .. note:: This is rate-limited to protect the network from flooding.
                  This is primarily useful for synchronizing network updates
                  with user code.
        
        .. versionadded:: 2017.0.0
        """
        cls._api.flush()
    
    @classmethod
    def setUpdateRate(cls, interval):
        """Sets the period of time between writes to the network. 
        
        WPILib's networktables and SmartDashboard default to 100ms, we have
        set it to 50ms instead for quicker response time. You should not set
        this value too low, as it could potentially increase the volume of
        data sent over the network.
        
        :param interval: Write flush period in seconds (default is 0.050,
                         or 50ms)
        :type interval: float
        
        .. warning:: If you don't know what this setting affects, don't mess
                     with it!
                        
        .. versionadded:: 2017.0.0
        """
        cls._api.setUpdateRate(interval)

    # Deprecated alias
    setWriteFlushPeriod = setUpdateRate
    
    @classmethod
    def savePersistent(cls, filename):
        """Saves persistent keys to a file. The server does this automatically.
        
        :param filename: Name of file to save keys to
        :type filename: str
        
        :returns: None if success, or a string describing the error on failure
        
        .. versionadded:: 2017.0.0
        """
        return cls._api.savePersistent(filename)
        
    @classmethod
    def loadPersistent(cls, filename):
        """Loads persistent keys from a file. WPILib will do this automatically
        on a robot server.
        
        :param filename: Name of file to load keys from
        :type filename: str
        
        :returns: None if success, or a string describing the error on failure
        
        .. versionadded:: 2017.0.0
        """
        return cls._api.loadPersistent(filename)
    
    
    @classmethod
    def setDashboardMode(cls):
        """This will allow the driver station to connect to your code and
        receive the IP address of the robot from it. You must not call
        :meth:`setClientMode`, :meth:`setTeam`, or :meth:`setIPAddress`
        
        .. warning:: Only use this if your pynetworktables client is running
                     on the same host as the driver station, or nothing will
                     happen! 
                     
                     This mode will only connect to the robot if the FRC
                     Driver Station is able to connect to the robot and the
                     LabVIEW dashboard has been disabled.
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        with cls._staticMutex:
            cls._checkInit()
            cls._mode = 'dashboard'
            
    @classmethod
    def setTestMode(cls, server=True):
        """Setup network tables to run in unit test mode, and enables verbose
        logging.
        
        .. warning:: This must be called before :meth:`initalize` or :meth:`getTable`
        """
        with cls._staticMutex:
            cls._checkInit()
            if server:
                cls._mode = 'test-server'
            else:
                cls._mode = 'test-client'
            cls.enableVerboseLogging()
            
    @classmethod
    def enableVerboseLogging(cls):
        """Enable verbose logging that can be useful when trying to diagnose
        NetworkTables issues.
        
        .. warning:: Don't enable this in normal use, as it can potentially 
                     cause performance issues due to the volume of logging.
                  
        .. versionadded:: 2017.0.0
        """
        with cls._staticMutex:
            cls._checkInit()
            cls._api.setVerboseLogging(True)
    
    @classmethod
    def getTable(cls, key):
        """Gets the table with the specified key. If the table does not exist,
        a new table will be created.

        This will automatically initialize network tables if it has not been
        already initialized.

        :param key: the key name
        :type key: str
        
        :returns: the network table requested
        :rtype: :class:`.NetworkTable`
        """
        
        # Must start with separator
        if key.startswith(cls.PATH_SEPARATOR):
            path = key
        else:
            path = cls.PATH_SEPARATOR + key
        
        # Must not end with separator
        if path.endswith(cls.PATH_SEPARATOR):
            path = path[:-1]
        
        with cls._staticMutex:
            if not cls._running:
                cls.initialize()
            
            table = cls._tables.get(path)
            if table is not None:
                return table
            
            # TODO: circular import problem, fix in 2018
            from .networktable import NetworkTable
                
            table = NetworkTable(path, cls._api)
            
            # global table is special
            if path == '':
                table._path = ''
            
            cls._tables[path] = table
            return table

    @classmethod
    def getGlobalTable(cls):
        """Returns an object that allows you to write values to raw network table
        keys (which are paths with / separators).

        This will automatically initialize network tables if it has not been
        already.

        .. note:: This is now an alias for ``NetworkTables.getTable('/')``
        
        .. versionadded:: 2015.2.0
        .. versionchanged:: 2017.0.0
           Returns a NetworkTable instance

        :rtype: :class:`.NetworkTable`
        """
        return cls.getTable('/')

    @classmethod
    def addGlobalListener(cls, listener, immediateNotify=True,
                          localNotify=True):
        '''Adds a listener that will be notified when any key in any
        NetworkTable is changed. The keys that are received using this
        listener will be full NetworkTable keys. Most users will not
        want to use this listener type.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        This will automatically initialize network tables if it has not been
        already.

        :param listener: A callable that has this signature: `callable(key, value, isNew)`
        :param immediateNotify: If True, the listener will be called immediately with the current values of the table
        :param immediateNotify: bool

        .. versionadded:: 2015.2.0
        
        .. versionchanged:: 2017.0.0
           `paramIsNew` parameter added

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended as we are not
                     currently sure if deadlocks will occur
        '''
        gtable = cls.getTable('/')
        gtable.addTableListener(listener, immediateNotify=immediateNotify,
                                key=0xdeadbeef, localNotify=localNotify)
        
    @classmethod
    def addGlobalListenerEx(cls, listener, flags, paramIsNew=True):
        '''Adds a listener that will be notified when any key in any
        NetworkTable is changed. The keys that are received using this
        listener will be full NetworkTable keys. Most users will not
        want to use this listener type.
        
        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        This will automatically initialize network tables if it has not been
        already.
        
        :param listener: A callable that has this signature: `callable(key, value, isNew)`
        :param flags: Bitmask of flags that indicate the types of notifications you wish to receive
        :type flags: :class:`.NotifyFlags`
        :param paramIsNew: If True, the listener third parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags
        :type paramIsNew: bool
        '''
        gtable = cls.getTable('/')
        gtable.addTableListenerEx(listener, flags, key=0xdeadbeef,
                                  paramIsNew=paramIsNew)

    @classmethod
    def removeGlobalListener(cls, listener):
        '''Removes a global listener

        .. versionadded:: 2015.2.0
        '''
        cls.getTable('/').removeTableListener(listener)
            
    @classmethod
    def getGlobalAutoUpdateValue(cls, key, defaultValue, writeDefault):
        '''Global version of getAutoUpdateValue. This function will not initialize
        NetworkTables.
        
        :param key: the full NT path of the value (must start with /)
        :type key: str
        :param defaultValue: The default value to return if the key doesn't exist
        :type defaultValue: any
        :param writeDefault: If True, force the value to the specified default
        :type writeDefault: bool
        
        .. versionadded:: 2015.3.0
        
        .. seealso:: :func:`.ntproperty` is a read-write alternative to this
        '''
        assert key.startswith('/')
        
        # Use raw NT api to avoid having to initialize networktables
        value = None
        valuefn = None # optimization for ntproperty
        
        if not writeDefault:
            value = cls._api.getEntryValue(key)
            
        if value is None:
            valuefn = Value.getFactory(defaultValue)
            cls._api.setEntryValue(key, valuefn(defaultValue))
            value = defaultValue
        elif not writeDefault:
            value = value.value
            valuefn = Value.getFactory(value)
        else:
            valuefn = Value.getFactory(value)
        
        return cls._api.createAutoValue(key,
                                        AutoUpdateValue(key, value, valuefn))

    @classmethod
    def getRemoteAddress(cls):
        '''
            Only returns a valid address if connected to the server. If
            this is a server, returns None
            
            :returns: IP address of server or None
            :rtype: str
            
            .. versionadded:: 2015.3.2
        '''
        return cls._api.getRemoteAddress()

    @classmethod
    def isConnected(cls):
        """
            :returns: True if connected to at least one other NetworkTables
                      instance
            :rtype: bool
        """
        return cls._api.getIsConnected()

    @classmethod
    def isServer(cls):
        """:returns: True if configured in server mode"""
        return 'server' in cls._mode
    
    @classmethod
    def addConnectionListener(cls, listener, immediateNotify=False):
        '''Adds a listener that will be notified when a new connection to a 
        NetworkTables client/server is established.
        
        The listener is called from a NetworkTables owned thread and should
        return as quickly as possible.
        
        :param listener: A function that will be called with two parameters
        :type listener: fn(bool, ConnectionInfo)
        
        :param immediateNotify: If True, the listener will be called immediately
                                with any active connection information
        :type immediateNotify: bool
        
        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended.
                     
        .. versionchanged:: 2017.0.0
           The listener is now a function
        '''
        assert callable(listener)
        cls._api.addConnectionListener(listener, immediateNotify)

    @classmethod
    def removeConnectionListener(cls, listener):
        '''Removes a connection listener
        
        :param listener: The function registered for connection notifications
        '''
        cls._api.removeConnectionListener(listener)




# validated: 2016-10-27 DS a73166a src/Notifier.cpp src/Notifier.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2015. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import namedtuple
import threading

from .constants import (
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS
)

from .support.compat import Queue

import logging
logger = logging.getLogger('nt')


_EntryCallback = namedtuple('EntryCallback', [
    'prefix',
    'callback',
    'flags'
])

_EntryNotification = namedtuple('EntryNotification', [
    'is_entry',
    'name',
    'value',
    'flags',
    'only'
])

_ConnectionNotification = namedtuple('ConnectionNotification', [
    'is_entry',
    'connected',
    'conn_info',
    'only'
])

class _Escape(Exception):
    pass


class UidVector(dict):
    
    def __init__(self):
        self.idx = 0
        self.lock = threading.Lock()
    
    def add(self, item):
        with self.lock:
            idx = self.idx
            self.idx += 1
        
        self[idx] = item
        return idx


_assign_both = NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS


class Notifier(object):

    def __init__(self, verbose=False):
        self.m_mutex = threading.Lock()
        
        self.m_verbose = verbose
        
        self.m_active = False
        self.m_owner = None
        
        self.m_local_notifiers = False
        self.m_local_entry_notifiers = False
        
        self.m_entry_listeners = UidVector()
        self.m_conn_listeners = set()
        
        # In python we don't need multiple queues
        self.m_notifications = Queue()
        
        self.m_on_start = None
        self.m_on_exit = None
        
        # Python specific: autoValue support
        self.m_autovalues = {}
        
    def setVerboseLogging(self, verbose):
        self.m_verbose = verbose
    
    def start(self):
        if not self.m_owner:
            self.m_active = True
            self.m_owner = threading.Thread(target=self._thread_main, name='notifier_thread')
            self.m_owner.daemon = True
            self.m_owner.start()
    
    def stop(self):
        if self.m_owner:
            self.m_active = False
            self.m_notifications.put(None)
            self.m_owner.join()
            self.m_owner = None
    
    def _thread_main(self):
        
        if self.m_on_start:
            self.m_on_start()
        
        try:
            while self.m_active:
                item = self.m_notifications.get()
                    
                if not self.m_active:
                    raise _Escape() # goto done
        
                # Entry notifications
                if item.is_entry:
        
                    if not item.value:
                        continue
                    
                    item_name = item.name
                    item_value = item.value.value
                    item_flags = item.flags
        
                    if item.only:
                        try:
                            # ntcore difference: no uid in callback
                            item.only(item_name, item_value, item_flags)
                        except Exception:
                            logger.warn("Unhandled exception processing notify callback", exc_info=True)
                        continue
                    
                    # Use copy because iterator might get invalidated.
                    for listener in list(self.m_entry_listeners.values()):
                        
                        # Flags must be within requested flag set for this listener.
                        # Because assign messages can result in both a value and flags update,
                        # we handle that case specially.
                        listen_flags = listener.flags
                        flags = item_flags
                        
                        if (flags & _assign_both) == _assign_both:
                            if (listen_flags & _assign_both) == 0:
                                continue
        
                            listen_flags &= ~_assign_both
                            flags &= ~_assign_both
        
                        if (flags & ~listen_flags) != 0:
                            continue
                        
                        # must match prefix
                        if not item_name.startswith(listener.prefix):
                            continue
                        
                        try:
                            # ntcore difference: no uid in callback
                            listener.callback(item_name, item_value, item_flags)
                        except Exception:
                            logger.warn("Unhandled exception processing notify callback", exc_info=True)
                
                # Connection notifications
                else:
                    if item.only:
                        try:
                            # ntcore difference: no uid in callback
                            item.only(item.connected, item.conn_info)
                        except Exception:
                            logger.warn("Unhandled exception processing notify callback", exc_info=True)
                        continue
                    
                    # Use copy because iterator might get invalidated.
                    for listener in list(self.m_conn_listeners):
                        try:
                            # ntcore difference: no uid in callback
                            listener(item.connected, item.conn_info)
                        except Exception:
                            logger.warn("Unhandled exception processing notify callback", exc_info=True)
        
        except _Escape:
            pass # because goto doesn't exist in python
        except Exception:
            logger.exception("Unhandled exception in notifier thread")
        
        logger.debug('Notifier thread exiting')
        
        if self.m_on_exit:
            self.m_on_exit()
    
    
    def addEntryListener(self, prefix, callback, flags):
        if self.m_verbose:
            logger.debug("%s entry listeners active", len(self.m_entry_listeners) + 1)
        
        self.start()
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
            self.m_local_entry_notifiers = True
    
        return self.m_entry_listeners.add(_EntryCallback(prefix, callback, flags))
    
    def removeEntryListener(self, entry_listener_uid):
        try:
            del self.m_entry_listeners[entry_listener_uid]
        except KeyError:
            pass
        
    def createAutoValue(self, key, value):
        self.m_local_notifiers = True
        return self.m_autovalues.setdefault(key, value)
    
    def notifyEntry(self, name, value, flags, only=None):
        
        # Python-specific autovalue support
        vv = value.value
        if vv is not None:
            autovalue = self.m_autovalues.get(name)
            if autovalue is not None:
                autovalue._AutoUpdateValue__value = value.value
        
        # optimization: don't generate needless local queue entries if we have
        # no local listeners (as this is a common case on the server side)
        if not self.m_local_entry_notifiers and (flags & NT_NOTIFY_LOCAL) != 0:
            return
    
        if not self.m_owner:
            return
        
        self.m_notifications.put(_EntryNotification(True, name, value, flags, only))
    
    
    def addConnectionListener(self, callback):
        if self.m_verbose:
            logger.debug("%s connection listeners active", len(self.m_conn_listeners) + 1)
        
        self.start()
        self.m_conn_listeners.add(callback)
    
    def removeConnectionListener(self, callback):
        try:
            self.m_conn_listeners.remove(callback)
        except KeyError:
            pass
    
    def notifyConnection(self, connected, conn_info, only=None):
        if not self.m_owner:
            return
        
        self.m_notifications.put(_ConnectionNotification(False, connected, conn_info, only))



# novalidate

import socket

from .tcp_stream import TCPStream

def TcpConnector(server, port, timeout):
    def _connect():
        
        if timeout == 0:
            sd = socket.create_connection((server, port))
        else:
            sd = socket.create_connection((server, port), timeout)
            sd.settimeout(None)
        
        return TCPStream(sd, server, port)

    return _connect


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# When running, this will continue incrementing the value 'dsTime', and the
# value should be visible to other networktables clients and the robot.
#

import sys
import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTables.initialize(server=ip)

sd = NetworkTables.getTable("/GRIP/myContoursReport")
#sna.append(1)

#sd.putValue('centerX', na)
sd.putStringArray('sau', (u'𐄜', '2017'))

i = 0

while True:
    try:
        pass
        #print('centerX:', ret)
    except KeyError:
        print('centerX: N/A')

    time.sleep(1)
    i += 1




# validated: 2016-10-27 DS a7eca7d src/RpcServer.cpp src/RpcServer.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2015. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import namedtuple

import threading

from .message import Message
from .support.compat import Queue, Empty

import logging
logger = logging.getLogger('nt')


RpcCall = namedtuple('RpcCall', [
    'name',
    'msg',
    'func',
    'conn_id',
    'send_response',
    'conn_info'
])


class RpcServer(object):
    
    def __init__(self):
        
        self.m_call_thread = None
        
        self.m_mutex = threading.Lock()
        
        self.m_poll_queue = Queue()
        self.m_response_map = {}
        
        self.m_terminating = False
        self.m_on_start = None
        self.m_on_exit = None
        
        self.m_call_queue = Queue()

    def setOnStart(self, on_start):
        self.m_on_start = on_start
        
    def setOnExit(self, on_exit):
        self.m_on_exit = on_exit
   
    def start(self):
        if not self.m_call_thread:
            self.m_call_thread = threading.Thread(target=self._callThread,
                                                  name='nt-rpc-thread')
            self.m_call_thread.daemon = True
            self.m_call_thread.start()
             
    def stop(self):
        if self.m_call_thread:
            self.m_terminating = True
            self.m_call_queue.put(None)
            
            self.m_call_thread.join(1)
            if self.m_call_thread.is_alive():
                logger.warn("%s did not die", self.m_call_thread.name)
    
    def processRpc(self, name, msg, func, conn_id, send_response, conn_info):
        call = RpcCall(name, msg, func, conn_id, send_response, conn_info)
        if func:
            if self.m_call_thread:   
                self.m_call_queue.put(call)
        else:
            self.m_poll_queue.put(call)
            self.m_poll_cond.notify_one()
    
    def pollRpc(self, blocking, call_info, time_out=None):
        
        item = None
        while item is None:
            if self.m_terminating:
                return False
            
            try:
                item = self.m_poll_queue.get(blocking, time_out)
            except Empty:
                return False
    
        # do not include conn id if the result came from the server
        if item.conn_id != 0xffff:
            call_uid = (item.conn_id << 16) | item.msg.seq_num_uid()
        else:
            call_uid = item.msg.seq_num_uid()
    
        call_info.rpc_id = item.msg.id()
        call_info.call_uid = call_uid
        call_info.name = item.name
        call_info.params = item.msg.str()
        self.m_response_map[(item.msg.id(), call_uid)] = item.send_response
        
        return True
    
    def postRpcResponse(self, rpc_id, call_uid, result):
        send_response = self.m_response_map.pop((rpc_id, call_uid), None)
        if send_response is None:
            logger.warn("posting RPC response to nonexistent call (or duplicate response)")
            return
    
        send_response(Message.rpcResponse(rpc_id, call_uid, result))
    
    def _callThread(self):
        if self.m_on_start:
            self.m_on_start()
        
        while not self.m_terminating:
            
            item = self.m_call_queue.get()
            if not item:
                continue
            
            logger.debug("rpc calling %s", item.name)

            if not item.name or not item.msg or not item.func or not item.send_response:
                continue
            
            try:
                result = item.func(item.name, item.msg.str(), item.conn_info)
            except Exception:
                logger.warn("Exception while executing callback", exc_info=1)
            else:
                item.send_response(Message.rpcResponse(item.msg.id(),
                                                        item.msg.seq_num_uid(), result))
        
        if self.m_on_exit:
            self.m_on_exit()



# validated: 2017-01-07 DS 95e5295666c6 wpiutil/src/tcpsockets/TCPStream.cpp

import select
import socket
import threading

class StreamEOF(IOError):
    pass

class TCPStream(object):

    def __init__(self, sd, peer_ip, peer_port):
        
        self.m_sd = sd
        self.m_peerIP = peer_ip
        self.m_peerPort = peer_port
        
        self.m_rdsock = sd.makefile('rb')
        self.m_wrsock = sd.makefile('wb')
        
        self.close_lock = threading.Lock()
    
    def read(self, size):
        
        # TODO: ntcore does a select to wait for read to be available. Necessary?
        
        data = self.m_rdsock.read(size)
        if size > 0 and len(data) != size:
            raise StreamEOF("end of file")
        return data

    def readStruct(self, s):
        sz = s.size
        data = self.m_rdsock.read(sz)
        if len(data) != sz:
            raise StreamEOF("end of file")
        return s.unpack(data)
    
    def send(self, contents):
        self.m_wrsock.write(contents)
        self.m_wrsock.flush()
    
    def close(self):
        with self.close_lock:
            if self.m_sd:
                try:
                    self.m_sd.shutdown(socket.SHUT_RDWR)
                except OSError:
                    pass
                
                self.m_sd.close()
                self.m_sd = None
    
    def getPeerIP(self):
        return self.m_peerIP
    
    def getPeerPort(self):
        return self.m_peerPort
    
    def setNoDelay(self):
        self.m_sd.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    
    def _waitForReadEvent(self, timeout):
        r, _, _ = select.select((self.m_sd,),(),(), timeout)
        return len(r) > 0



#!/usr/bin/env python3
#
# This is a NetworkTables server (eg, the robot or simulator side).
#
# On a real robot, you probably would create an instance of the 
# wpilib.SmartDashboard object and use that instead -- but it's really
# just a passthru to the underlying NetworkTable object.
#
# When running, this will continue incrementing the value 'robotTime',
# and the value should be visible to networktables clients such as 
# SmartDashboard. To view using the SmartDashboard, you can launch it
# like so:
#
#     SmartDashboard.jar ip 127.0.0.1
#

import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

NetworkTables.enableVerboseLogging()
table = NetworkTables.getGlobalTable()

try:

    i = 0
    while True:
        try:
            print('dsTime:', table.getNumber('/SmartDashboard/dsTime'))
        except KeyError:
            print('dsTime: N/A')
    
        table.putNumber('/SmartDashboard/robotTime', i)
        time.sleep(1)
        i += 1

finally:
    NetworkTables.shutdown()



__all__ = ["NetworkTable"]

# TODO: in 2018, remove this circular import
from .networktables import NetworkTables as _NT

from ntcore.constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    
    NT_PERSISTENT,
    
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS
)

_is_new = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW

from ntcore.value import Value

import logging
logger = logging.getLogger('nt')


class _defaultValueSentry:
    pass

class NetworkTable:
    '''
        This is a NetworkTable instance, it allows you to interact with
        NetworkTables in a table-based manner. You should not directly
        create a NetworkTable object, but instead use
        :meth:`.NetworkTables.getTable` to retrieve a NetworkTable instance.
        
        For example, to interact with the SmartDashboard::

            from networktables import NetworkTables
            sd = NetworkTables.getTable('SmartDashboard')
    
            sd.putNumber('someNumber', 1234)
            ...
            
        .. seealso::
           - The examples in the documentation.
           - :class:`.NetworkTables`
    '''
    
    PATH_SEPARATOR = '/'
    
    # These static aliases are deprecated and will be removed in 2018!
    
    initialize = _NT.initialize
    shutdown = _NT.shutdown
    setClientMode = _NT.setClientMode
    setServerMode = _NT.setServerMode
    setTeam = _NT.setTeam
    setIPAddress = _NT.setIPAddress
    setPort = _NT.setPort
    setPersistentFilename = _NT.setPersistentFilename
    setNetworkIdentity = _NT.setNetworkIdentity
    globalDeleteAll = _NT.globalDeleteAll
    flush = _NT.flush
    setUpdateRate = _NT.setUpdateRate
    setWriteFlushPeriod = _NT.setWriteFlushPeriod
    savePersistent = _NT.savePersistent
    loadPersistent = _NT.loadPersistent
    setDashboardMode = _NT.setDashboardMode
    setTestMode = _NT.setTestMode
    getTable = _NT.getTable
    getGlobalTable = _NT.getGlobalTable
    addGlobalListener = _NT.addGlobalListener
    removeGlobalListener = _NT.removeGlobalListener
    getGlobalAutoUpdateValue = _NT.getGlobalAutoUpdateValue
    
    addConnectionListener = _NT.addConnectionListener
    removeConnectionListener = _NT.removeConnectionListener
    
    getRemoteAddress = _NT.getRemoteAddress
    isConnected = _NT.isConnected
    isServer = _NT.isServer
    
    
    def __init__(self, path, api):
        
        #: Path of table without trailing slash
        self.path = path
        self._path = path + self.PATH_SEPARATOR
        self._pathsz = len(self._path)
            
        self._api = api
        
        self._listeners = {}

    def __str__(self):
        return "NetworkTable: "+self._path
    
    def __repr__(self):
        return "<NetworkTable path=%s>" % self._path

    def addTableListener(self, listener, immediateNotify=False, key=None,
                               localNotify=False):
        '''Adds a listener that will be notified when any key in this
        NetworkTable is changed, or when a specified key changes.
        
        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.
        
        :param listener: A callable with signature `callable(source, key, value, isNew)`
        :param immediateNotify: If True, the listener will be called immediately with the current values of the table
        :type immediateNotify: bool
        :param key: If specified, the listener will only be called when this key is changed
        :type key: str
        :param localNotify: True if you wish to be notified of changes made locally (default is False)
        :type localNotify: bool
        
        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended
                     
        .. versionchanged:: 2017.0.0
           Added localNotify parameter (defaults to False, which is different from NT2)
        
        '''
        flags = NT_NOTIFY_NEW | NT_NOTIFY_UPDATE
        if immediateNotify:
            flags |= NT_NOTIFY_IMMEDIATE
        if localNotify:
            flags |= NT_NOTIFY_LOCAL
            
        self.addTableListenerEx(listener, flags, key=key)
        
    def addTableListenerEx(self, listener, flags, key=None,
                           paramIsNew=True):
        '''Adds a listener that will be notified when any key in this
        NetworkTable is changed, or when a specified key changes.
        
        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.
        
        :param listener: A callable with signature `callable(source, key, value, param)`
        :param flags: Bitmask of flags that indicate the types of notifications you wish to receive
        :type flags: :class:`.NotifyFlags`
        :param key: If specified, the listener will only be called when this key is changed
        :type key: str
        :param paramIsNew: If True, the listener fourth parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags
        :type paramIsNew: bool
        
        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended
        
        .. versionadded:: 2017.0.0
        '''
        
        if key is None:
            _pathsz = self._pathsz
            if paramIsNew:
                def callback(key_, value_, flags_):
                    key_ = key_[_pathsz:]
                    if '/' not in key_:
                        listener(self, key_, value_, (flags_ & _is_new) != 0)
            else:
                def callback(key_, value_, flags_):
                    key_ = key_[_pathsz:]
                    if '/' not in key_:
                        listener(self, key_, value_, flags_)
    
        # Hack: Internal flag used by addGlobalListener*
        elif key == 0xdeadbeef:
            if paramIsNew:
                def callback(key_, value_, flags_):
                    listener(key_, value_, (flags_ & _is_new) != 0)
            else:
                callback = listener
                
        else:
            path = self._path + key
            if paramIsNew:
                def callback(key_, value_, flags_):
                    if path == key_:
                        listener(self, key, value_, (flags_ & _is_new) != 0)
            else:
                def callback(key_, value_, flags_):
                    if path == key_:
                        listener(self, key, value_, flags_)
        
        uid = self._api.addEntryListener(self._path, callback, flags)
        self._listeners.setdefault(listener, []).append(uid)

    def addSubTableListener(self, listener, localNotify=False):
        '''Adds a listener that will be notified when any key in a subtable of
        this NetworkTable is changed.
        
        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.
        
        :param listener: Callable to call when previously unseen table appears.
                         Function signature is `callable(source, key, subtable, True)`
        :param localNotify: True if you wish to be notified when local changes
                            result in a new table
        :type localNotify: bool
        
        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended as we are not
                     currently sure if deadlocks will occur
                     
        .. versionchanged:: 2017.0.0
           Added localNotify parameter
        '''
        notified_tables = {}
        
        def _callback(key, value, _):
            key = key[self._pathsz:]
            if '/' in key:
                skey = key[:key.index('/')]
                
                o = object()
                if notified_tables.setdefault(skey, o) is o:
                    try:
                        listener(self, skey, self.getSubTable(skey), True)
                    except Exception:
                        logger.warning("Unhandled exception in %s", listener, exc_info=True)
        
        flags = NT_NOTIFY_NEW | NT_NOTIFY_IMMEDIATE
        if localNotify:
            flags |= NT_NOTIFY_LOCAL
        
        uid = self._api.addEntryListener(self._path, _callback, flags)
        self._listeners.setdefault(listener, []).append(uid)
        

    def removeTableListener(self, listener):
        '''Removes a table listener
        
        :param listener: callable that was passed to :meth:`.addTableListener`
                         or :meth:`.addSubTableListener`
        '''
        uids = self._listeners.pop(listener, [])
        for uid in uids:
            self._api.removeEntryListener(uid)

    def getSubTable(self, key):
        """Returns the table at the specified key. If there is no table at the
        specified key, it will create a new table

        :param key: the key name
        :type key: str
        
        :returns: the networktable to be returned
        :rtype: :class:`.NetworkTable`
        """
        path = self._path + key
        return _NT.getTable(path)
        

    def containsKey(self, key):
        """Determines whether the given key is in this table.
        
        :param key: the key to search for
        :type key: str
        
        :returns: True if the table as a value assigned to the given key
        :rtype: bool
        """
        path = self._path + key
        return self._api.getEntryValue(path) is not None

    def __contains__(self, key):
        return self.containsKey(key)

    def containsSubTable(self, key):
        """Determines whether there exists a non-empty subtable for this key
        in this table.
        
        :param key: the key to search for (must not end with path separator)
        :type key: str
        
        :returns: True if there is a subtable with the key which contains at least
                  one key/subtable of its own
        :rtype: bool
        """
        path = self._path + key + self.PATH_SEPARATOR
        return len(self._api.getEntryInfo(path, 0)) > 0
    
    def getKeys(self, types=0):
        """
        :param types: bitmask of types; 0 is treated as a "don't care".
        :type types: :class:`.EntryTypes`
        
        :returns: keys currently in the table
        :rtype: list
        
        .. versionadded:: 2017.0.0
        """
        keys = []
        for entry in self._api.getEntryInfo(self._path, types):
            relative_key = entry.name[len(self._path):]
            if self.PATH_SEPARATOR in relative_key:
                continue
            
            keys.append(relative_key)
        
        return keys
        
    def getSubTables(self):
        """:returns: subtables currently in the table
        :rtype: list
        
        .. versionadded:: 2017.0.0
        """
        keys = set()
        for entry in self._api.getEntryInfo(self._path, 0):
            relative_key = entry.name[len(self._path):]
            subst = relative_key.split(self.PATH_SEPARATOR)
            if len(subst) == 1:
                continue
            
            keys.add(subst[0])
        
        return list(keys)
    
    def setPersistent(self, key):
        """Makes a key's value persistent through program restarts.
        
        :param key: the key to make persistent
        :type key: str
        
        .. versionadded:: 2017.0.0
        """
        self.setFlags(key, NT_PERSISTENT)
        
    def clearPersistent(self, key):
        """Stop making a key's value persistent through program restarts.
        
        :param key: the key name
        :type key: str
        
        .. versionadded:: 2017.0.0
        """
        self.clearFlags(key, NT_PERSISTENT)
    
    def isPersistent(self, key):
        """Returns whether the value is persistent through program restarts.
        
        :param key: the key name
        :type key: str
        
        .. versionadded:: 2017.0.0
        """
        return self.getFlags(key) & NT_PERSISTENT != 0
        
    def delete(self, key):
        """Deletes the specified key in this table.
        
        :param key: the key name
        :type key: str
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.deleteEntry(path)
    
    def setFlags(self, key, flags):
        """Sets entry flags on the specified key in this table.
        
        :param key: the key name
        :type key: str
        :param flags: the flags to set (bitmask)
        :type flags: :class:`.EntryFlags`
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.setEntryFlags(path, self._api.getEntryFlags(path) | flags)
        
    def clearFlags(self, key, flags):
        """Clears entry flags on the specified key in this table.
        
        :param key: the key name
        :type key: str
        :param flags: the flags to clear (bitmask)
        :type flags: :class:`.EntryFlags`
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.setEntryFlags(path, self._api.getEntryFlags(path) & ~flags)
        
    def getFlags(self, key):
        """Returns the entry flags for the specified key.
        
        :param key: the key name
        :type key: str
        :returns: the flags, or 0 if the key is not defined
        :rtype: :class:`.EntryFlags`
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.getEntryFlags(path)

    def putNumber(self, key, value):
        """Put a number in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: int, float
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeDouble(value))

    def setDefaultNumber(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: int, float
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeDouble(defaultValue))

    def getNumber(self, key, defaultValue=_defaultValueSentry):
        """Gets the number associated with the given name.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: int, float
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: int, float
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_DOUBLE:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value

    def putString(self, key, value):
        """Put a string in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: str
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeString(value))

    def setDefaultString(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: str
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeString(defaultValue))

    def getString(self, key, defaultValue=_defaultValueSentry):
        """Gets the string associated with the given name. If the key does not
        exist or is of different type, it will return the default value.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: str
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: str
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_STRING:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value

    def putBoolean(self, key, value):
        """Put a boolean in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: bool
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeBoolean(value))

    def setDefaultBoolean(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: bool
        
        :returns: False if the table key exists with a different type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeBoolean(defaultValue))

    def getBoolean(self, key, defaultValue=_defaultValueSentry):
        """Gets the boolean associated with the given name. If the key does not
        exist or is of different type, it will return the default value.

        :param key: the key name
        :type key: str
        :param defaultValue: the default value if the key is None.  If not
                             specified, raises KeyError if the key is None.
        :type defaultValue: bool
        
        :returns: the key
        :rtype: bool
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_BOOLEAN:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value

    def putBooleanArray(self, key, value):
        """Put a boolean array in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: list(bool)
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeBooleanArray(value))
    
    def setDefaultBooleanArray(self, key, defaultValue=_defaultValueSentry):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: list(bool)
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeBooleanArray(defaultValue))
        
    def getBooleanArray(self, key, defaultValue=_defaultValueSentry):
        """Returns the boolean array the key maps to. If the key does not exist or is
        of different type, it will return the default value.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: list(bool)
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: list(bool)
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_BOOLEAN_ARRAY:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value
    
    def putNumberArray(self, key, value):
        """Put a number array in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: list(float)
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeDoubleArray(value))
    
    def setDefaultNumberArray(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: list(int, float)
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeDoubleArray(defaultValue))
    
    def getNumberArray(self, key, defaultValue=_defaultValueSentry):
        """Returns the number array the key maps to. If the key does not exist or is
        of different type, it will return the default value.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: list(int, float)
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: list(int, float)
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_DOUBLE_ARRAY:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value
        
    def putStringArray(self, key, value):
        """Put a string array in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: list(str)
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeStringArray(value))
    
    def setDefaultStringArray(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: list(str)
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeStringArray(defaultValue))
    
    def getStringArray(self, key, defaultValue=_defaultValueSentry):
        """Returns the string array the key maps to. If the key does not exist or is
        of different type, it will return the default value.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: list(str)
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: list(str)
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_STRING_ARRAY:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value
        
    def putRaw(self, key, value):
        """Put a raw value (byte array) in the table
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: bytes
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeRaw(value))
        
    def setDefaultRaw(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: bytes
        
        :returns: False if the table key exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeRaw(defaultValue))
    
    def getRaw(self, key, defaultValue=_defaultValueSentry):
        """Returns the raw value (byte array) the key maps to. If the key does not
        exist or is of different type, it will return the default value.
        
        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found
        :type defaultValue: bytes
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        :rtype: bytes
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_RAW:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value
    
    def putValue(self, key, value):
        """Put a value in the table, trying to autodetect the NT type of
        the value. Refer to this table to determine the type mapping:
        
        ======= ============================ =================================
        PyType  NT Type                       Notes
        ======= ============================ =================================
        bool    :attr:`.EntryTypes.BOOLEAN`
        int     :attr:`.EntryTypes.DOUBLE`
        float   :attr:`.EntryTypes.DOUBLE`
        str     :attr:`.EntryTypes.STRING`
        bytes   :attr:`.EntryTypes.RAW`      Doesn't work in Python 2.7
        list    Error                        Use `putXXXArray` methods instead
        tuple   Error                        Use `putXXXArray` methods instead
        ======= ============================ =================================
        
        :param key: the key to be assigned to
        :type key: str
        :param value: the value that will be assigned
        :type value: bool, int, float, str, bytes
        
        :returns: False if the table key already exists with a different type
        :rtype: bool
        
        .. versionadded:: 2017.0.0
        """
        value = Value.getFactory(value)(value)
        path = self._path + key
        return self._api.setEntryValue(path, value)

    def setDefaultValue(self, key, defaultValue):
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :type key: str
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: bool, int, float, str, bytes
        
        :returns: False if the table key exists with a different type
        
        .. versionadded:: 2017.0.0
        
        .. seealso:: :meth:`.putValue`
        """
        defaultValue = Value.getFactory(defaultValue)(defaultValue)
        path = self._path + key
        return self._api.setDefaultEntryValue(path, defaultValue)

    def getValue(self, key, defaultValue=_defaultValueSentry):
        """Gets the value associated with a key. This supports all
        NetworkTables types (unlike :meth:`putValue`).
        
        :param key: the key of the value to look up
        :type key: str
        :param defaultValue: The default value to return if the key doesn't exist
        :type defaultValue: any
        
        :returns: the value associated with the given key
        :rtype: bool, int, float, str, bytes, list
        
        :raises KeyError: If the value doesn't exist and no default is provided, or
                          if it is the wrong type
        
        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value:
            if defaultValue != _defaultValueSentry:
                return defaultValue
            raise KeyError(path)

        return value.value
    
    def getAutoUpdateValue(self, key, defaultValue, writeDefault=True):
        '''Returns an object that will be automatically updated when the
        value is updated by networktables.
        
        :param key: the key name
        :type  key: str
        :param defaultValue: Default value to use if not in the table
        :type  defaultValue: any
        :param writeDefault: If True, put the default value to the table,
                             overwriting existing values
        :type  writeDefault: bool
        
        :rtype: :class:`.AutoUpdateValue`
        
        .. note:: If you modify the returned value, the value will NOT
                  be written back to NetworkTables. See :func:`.ntproperty`
                  if you're looking for that sort of thing.
        
        .. seealso:: :func:`.ntproperty` is a better alternative to use
        
        .. versionadded:: 2015.1.3
        '''
        return _NT.getGlobalAutoUpdateValue(self._path + key, defaultValue, writeDefault)

    





class AutoUpdateValue:
    """Holds a value from NetworkTables, and changes it as new entries
    come in. Updates to this value are NOT passed on to NetworkTables.
    
    Do not create this object directly, as it only holds the value. 
    Use :meth:`.NetworkTables.getAutoUpdateValue` to obtain an instance
    of this.
    """
    
    __slots__ = ['key', '__value', '_valuefn']
    
    def __init__(self, key, default, valuefn):
        self.key = key
        self.__value = default
        self._valuefn = valuefn
        
    def get(self):
        '''Returns the value held by this object'''
        return self.__value
    
    @property
    def value(self):
        return self.__value
    
    # Comparison operators et al
    
    def __lt__(self, other):
        raise TypeError("< not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    def __le__(self, other):
        raise TypeError("<= not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    def __eq__(self, other):
        raise TypeError("== not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    def __ne__(self, other):
        raise TypeError("!= not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    def __gt__(self, other):
        raise TypeError("> not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    def __ge__(self, other):
        raise TypeError(">= not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    
    def __bool__(self):
        raise TypeError("< not allowed on AutoUpdateValue objects. Use the .value attribute instead")
    
    def __hash__(self):
        raise TypeError("__hash__ not allowed on AutoUpdateValue objects")
    
    def __repr__(self):
        return '<AutoUpdateValue: %s>' % (self.__value.__repr__(), )




# validated: 2016-10-27 DS 273a395 src/Dispatcher.cpp
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2015. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import threading
import time

from .message import Message
from .network_connection import NetworkConnection

from .tcpsockets.tcp_acceptor import TcpAcceptor
from .tcpsockets.tcp_connector import TcpConnector

from .support.compat import monotonic, Condition

from .constants import (
    kKeepAlive,
    kClientHello,
    kProtoUnsup,
    kServerHello,
    kServerHelloDone,
    kClientHelloDone,
    kEntryAssign,
)

import logging
logger = logging.getLogger('nt')


class Dispatcher(object):
    
    def __init__(self, storage, notifier, verbose=False):
        
        # logging debugging
        self.m_verbose = verbose
        
        self.m_storage = storage
        self.m_notifier = notifier
        self.m_server = False
        self.m_persist_filename = None
        self.m_server_acceptor = None
        self.m_client_connectors = []
        self.m_identity = ''
        self.m_default_proto = 0x0300 # for testing
        
        # Mutex for user-accessible items
        self.m_user_mutex = threading.RLock()
        self.m_connections = []
        
        self.m_active = False # set to false to terminate threads
        self.m_update_rate = 0.050 # periodic dispatch rate, in s
        
        self.m_flush_mutex = threading.Lock()
        self.m_flush_cv = Condition(self.m_flush_mutex)
        self.m_last_flush = 0
        self.m_do_flush = False
        
        self.m_reconnect_cv = Condition(self.m_user_mutex)
        self.m_reconnect_proto_rev = self.m_default_proto
        self.m_do_reconnect = True
    
    def setVerboseLogging(self, verbose):
        self.m_verbose = verbose
    
    def startServer(self, persist_filename, listen_address, port):
        acceptor = TcpAcceptor(port, listen_address)
        self._startServer(persist_filename, acceptor)
    
    def startClient(self, servers):
        # servers is a tuple of (server, port)
        connectors = [TcpConnector(server, port, 1) for server, port in servers]
        self._startClient(connectors)
    
    def _startServer(self, persist_filename, acceptor):
        
        with self.m_user_mutex:
            if self.m_active:
                return
    
            self.m_active = True
    
        self.m_server = True
        self.m_persist_filename = persist_filename
        self.m_server_acceptor = acceptor
    
        # Load persistent file.  Ignore errors, pass along warnings.
        if persist_filename:
            self.m_storage.loadPersistent(persist_filename)
    
        self.m_storage.setOutgoing(self._queueOutgoing,
                                   self.m_server)
    
        self.m_dispatch_thread = threading.Thread(target=self._dispatchThreadMain,
                                                  name='nt-dispatch-thread') 
        self.m_clientserver_thread = threading.Thread(target=self._serverThreadMain,
                                                      name='nt-server-thread')
        
        self.m_dispatch_thread.daemon = True
        self.m_clientserver_thread.daemon = True
        
        self.m_dispatch_thread.start()
        self.m_clientserver_thread.start()
    
    def _startClient(self, connectors):
        if not isinstance(connectors, list):
            connectors = [connectors]
        
        with self.m_user_mutex:
            if self.m_active:
                return
    
            self.m_active = True
            self.m_client_connectors = connectors[:]
    
        self.m_server = False
        self.m_storage.setOutgoing(self._queueOutgoing,
                                   self.m_server)
    
        self.m_dispatch_thread = threading.Thread(target=self._dispatchThreadMain,
                                                  name='nt-dispatch-thread') 
        self.m_clientserver_thread = threading.Thread(target=self._clientThreadMain,
                                                      name='nt-client-thread')
        
        self.m_dispatch_thread.daemon = True
        self.m_clientserver_thread.daemon = True
        
        self.m_dispatch_thread.start()
        self.m_clientserver_thread.start()
    
    
    def stop(self):
        if not self.m_active:
            return
        
        self.m_active = False
    
        # wake up dispatch thread with a flush
        with self.m_flush_mutex:
            self.m_flush_cv.notify()
    
        # wake up client thread with a reconnect
        with self.m_user_mutex:
            del self.m_client_connectors[:]
    
        self._clientReconnect()
    
        # wake up server thread by shutting down the socket
        if self.m_server_acceptor:
            self.m_server_acceptor.shutdown()
        
        # join threads, timeout
        self.m_dispatch_thread.join(1)
        if self.m_dispatch_thread.is_alive():
            logger.warn("%s did not die", self.m_dispatch_thread.name)
        
        self.m_clientserver_thread.join(1)
        if self.m_clientserver_thread.is_alive():
            logger.warn("%s did not die", self.m_clientserver_thread.name)
        
        with self.m_user_mutex:
            conns = self.m_connections
            self.m_connections = []
    
        # close all connections
        for conn in conns:
            conn.stop()
            
        # cleanup the server socket
        if self.m_server_acceptor:
            self.m_server_acceptor.close()
    
    def setUpdateRate(self, interval):
        # don't allow update rates faster than 10 ms or slower than 1 second
        interval = float(interval)
        
        if interval < 0.01:
            interval = 0.01
    
        elif interval > 1.0:
            interval = 1.0
    
        self.m_update_rate = interval
    
    def setIdentity(self, name):
        with self.m_user_mutex:
            self.m_identity = name
    
    def setDefaultProtoRev(self, proto_rev):
        self.m_default_proto = proto_rev
        self.m_reconnect_proto_rev = proto_rev
    
    def flush(self):
        now = monotonic()
        with self.m_flush_mutex:
            # don't allow flushes more often than every 10 ms
            if (now - self.m_last_flush) < 0.010:
                return
    
            self.m_last_flush = now
            self.m_do_flush = True
    
            self.m_flush_cv.notify()
    
    def getConnections(self):
        conns = []
        if not self.m_active:
            return conns
    
        with self.m_user_mutex:
            for conn in self.m_connections:
                if conn.state() != NetworkConnection.State.kActive:
                    continue
        
                conns.append(conn.info())
        
        return conns
    
    def isConnected(self):
        if self.m_active:
            with self.m_user_mutex:
                for conn in self.m_connections:
                    if conn.is_connected():
                        return True
        return False
    
    def isServer(self):
        return self.m_server
    
    def notifyConnections(self, callback):
        with self.m_user_mutex:
            for conn in self.m_connections:
                conn.notifyIfActive(callback)
    
    def _dispatchWaitFor(self):
        return not self.m_active or self.m_do_flush
    
    def _dispatchThreadMain(self):
        timeout_time = monotonic()
    
        save_delta_time = 1.0
        next_save_time = timeout_time + save_delta_time
    
        count = 0
        is_server = self.m_server
        verbose = self.m_verbose
        
        while self.m_active:
            # handle loop taking too long
            start = monotonic()
            if start > timeout_time:
                timeout_time = start
    
            # wait for periodic or when flushed
            timeout_time += self.m_update_rate
            with self.m_flush_mutex:
                self.m_flush_cv.wait_for(self._dispatchWaitFor, timeout_time - start)
                self.m_do_flush = False
            
            # in case we were woken up to terminate
            if not self.m_active:
                break    
            
            # perform periodic persistent save
            if is_server and self.m_persist_filename and start > next_save_time:
                next_save_time += save_delta_time
                # handle loop taking too long
                if start > next_save_time:
                    next_save_time = start + save_delta_time
    
                err = self.m_storage.savePersistent(self.m_persist_filename, True)
                if err:
                    logger.warning("periodic persistent save: %s", err)
                
            with self.m_user_mutex:
                reconnect = False
    
                if verbose:
                    count += 1
                    if count > 10:
                        logger.debug("dispatch running %s connections",
                                     len(self.m_connections))
                        count = 0
                
                for conn in self.m_connections:
                    # post outgoing messages if connection is active
                    # only send keep-alives on client
                    state = conn.state()
                    if state == NetworkConnection.State.kActive:
                        conn.postOutgoing(not is_server)
                    
                    # if client, if connection died
                    if not is_server and state == NetworkConnection.State.kDead:
                        reconnect = True
                
                # reconnect if we disconnected (and a reconnect is not in progress)
                if reconnect and not self.m_do_reconnect:
                    self.m_do_reconnect = True
                    self.m_reconnect_cv.notify()
    
    def _queueOutgoing(self, msg, only, except_):
        with self.m_user_mutex:
            for conn in self.m_connections:
                if conn == except_:
                    continue
                
                if only and conn != only:
                    continue
                
                state = conn.state()
                if (state != NetworkConnection.State.kSynchronized and
                    state != NetworkConnection.State.kActive):
                    continue
    
                conn.queueOutgoing(msg)
    
    def _serverThreadMain(self):
        if not self.m_server_acceptor.start():
            self.m_active = False
        
        try:
            while self.m_active:
                stream = self.m_server_acceptor.accept()
                if not stream:
                    self.m_active = False
                    return
        
                if not self.m_active:
                    return
        
                logger.debug("server: client connection from %s port %s",
                             stream.getPeerIP(), stream.getPeerPort())
        
                # add to connections list
                conn = NetworkConnection(stream, self.m_notifier,
                                         self._serverHandshake,
                                         self.m_storage.getEntryType,
                                         verbose=self.m_verbose)
                
                conn.set_process_incoming(self.m_storage.processIncoming)
                    
                with self.m_user_mutex:
                    # reuse dead connection slots
                    for i in range(len(self.m_connections)):
                        c = self.m_connections[i]
                        if c.state() == NetworkConnection.State.kDead:
                            c.stop()
                            self.m_connections[i] = conn
                            break
                    else:
                        self.m_connections.append(conn)
        
                    conn.start()
        finally:
            logger.debug("server thread exiting")
    
    def _clientThreadMain(self):
        i = 0
        
        try:
            while self.m_active:
                # sleep between retries
                time.sleep(0.250)
        
                # get next server to connect to
                with self.m_user_mutex:
                    if not self.m_client_connectors:
                        continue
        
                    if i >= len(self.m_client_connectors):
                        i = 0
        
                    connect = self.m_client_connectors[i]
                    i += 1
                
                # try to connect (with timeout)
                if self.m_verbose:
                    logger.debug("client trying to connect")
                    
                try:
                    stream = connect()
                except IOError:
                    continue    # keep retrying
        
                logger.debug("client connected")
        
                with self.m_user_mutex:
                    conn = NetworkConnection(stream, self.m_notifier,
                                             self._clientHandshake,
                                             self.m_storage.getEntryType,
                                             verbose=self.m_verbose)
                    
                    conn.set_process_incoming(self.m_storage.processIncoming)
                    
                    # disconnect any current
                    for c in self.m_connections:
                        if c != conn:
                            c.stop()
                    
                    del self.m_connections[:]
                    self.m_connections.append(conn)
                    conn.set_proto_rev(self.m_reconnect_proto_rev)
                    conn.start()
                    
                    # reconnect the next time starting with latest protocol revision
                    self.m_reconnect_proto_rev = self.m_default_proto
                    
                    # block until told to reconnect
                    self.m_do_reconnect = False
                    
                    while not (not self.m_active or self.m_do_reconnect):
                        self.m_reconnect_cv.wait()
        finally:
            logger.debug("client thread exiting")
    
    def _clientHandshake(self, conn, get_msg, send_msgs):
        # get identity
        with self.m_user_mutex:
            self_id = self.m_identity
        
        # send client hello
        if self.m_verbose:
            logger.debug("client: sending hello")
        
        send_msgs((Message.clientHello(conn.get_proto_rev(), self_id),))
    
        # wait for response
        msg = get_msg()
        if not msg:
            # disconnected, retry
            logger.debug("client: server disconnected before first response")
            return False
        
        if msg.type == kProtoUnsup:
            if msg.id == 0x0200:
                logger.debug("client: connected to NT2 server, reconnecting...")
                self._clientReconnect(0x0200)
            else:
                logger.debug("client: connected to 0x%04x server, giving up...", msg.id)
    
            return False
        
        new_server = True
        if conn.get_proto_rev() >= 0x0300:
            # should be server hello; if not, disconnect.
            if not msg.type == kServerHello:
                return False
    
            remote_id = msg.str
            if (msg.flags & 1) != 0:
                new_server = False
    
            # get the next message
            msg = get_msg()
        else:
            remote_id = 'NT2 server'
            
        conn.set_remote_id(remote_id)
        
        # receive initial assignments
        incoming = []
        verbose = self.m_verbose
        
        while True:
            if not msg:
                # disconnected, retry
                logger.debug("client: server disconnected during initial entries")
                return False
    
            if msg.type == kServerHelloDone:
                break
            
            if not msg.type == kEntryAssign:
                # unexpected message
                logger.debug("client: received message (%s) other than entry assignment during initial handshake",
                             msg.type)
                return False
            
            if verbose:
                logger.debug("received assign str=%s id=%s seq_num=%s",
                             msg.str, msg.id, msg.seq_num_uid)
    
            incoming.append(msg)
            # get the next message
            msg = get_msg()
        
        # generate outgoing assignments
        outgoing = []
    
        self.m_storage.applyInitialAssignments(conn, incoming, new_server, outgoing)
    
        if conn.get_proto_rev() >= 0x0300:
            outgoing.append(Message.clientHelloDone())
        
        if outgoing:
            send_msgs(outgoing)
        
        return True
    
    def _serverHandshake(self, conn, get_msg, send_msgs):
        
        verbose = self.m_verbose
        
        # Wait for the client to send us a hello.
        msg = get_msg()
        if not msg:
            logger.debug("server: client disconnected before sending hello")
            return False
    
        if not msg.type == kClientHello:
            logger.debug("server: client initial message was not client hello")
            return False
    
        # Check that the client requested version is not too high.
        proto_rev = msg.id
        if proto_rev > self.m_default_proto:
            logger.debug("server: client requested proto > 0x%04x", self.m_default_proto)
            send_msgs((Message.protoUnsup(self.m_default_proto),))
            return False
    
        
        if proto_rev >= 0x0300:
            remote_id = msg.str
        else:
            remote_id = 'NT2 client'
        
        conn.set_remote_id(remote_id)
    
        # Set the proto version to the client requested version
        logger.debug("server: client protocol 0x%04x", proto_rev)
        conn.set_proto_rev(proto_rev)
    
        # Send initial set of assignments
        outgoing = []
    
        # Start with server hello.  TODO: initial connection flag
        if proto_rev >= 0x0300:
            with self.m_user_mutex:
                outgoing.append(Message.serverHello(0, self.m_identity))
        
        # Get snapshot of initial assignments
        self.m_storage.getInitialAssignments(conn, outgoing)
    
        # Finish with server hello done
        outgoing.append(Message.serverHelloDone())
    
        # Batch transmit
        if verbose:
            logger.debug("server: sending initial assignments")
        
        send_msgs(outgoing)
    
        # In proto rev 3.0 and later, handshake concludes with a client hello
        # done message, we can batch the assigns before marking the connection
        # active.  In pre-3.0, need to just immediately mark it active and hand
        # off control to the dispatcher to assign them as they arrive.
        if proto_rev >= 0x0300:
            # receive client initial assignments
            incoming = []
            while True:
                # get the next message (blocks)
                msg = get_msg()
                if not msg:
                    # disconnected, retry
                    logger.debug("server: disconnected waiting for initial entries")
                    return False
    
                if msg.type == kClientHelloDone:
                    break
                elif msg.type == kKeepAlive:
                    continue
                
                if msg.type != kEntryAssign:
                    # unexpected message
                    logger.debug("server: received message (%s) other than entry assignment during initial handshake",
                                 msg.type)
                    return False
                
                if verbose:
                    logger.debug("received assign str=%s id=%s seq_num=%s",
                                 msg.str, msg.id, msg.seq_num_uid)
    
                incoming.append(msg)
    
            for msg in incoming:
                self.m_storage.processIncoming(msg, conn)
    
        stream = conn.get_stream()
        return True
    
    def _clientReconnect(self, proto_rev=0x0300):
        if self.m_server:
            return
    
        with self.m_user_mutex:
            self.m_reconnect_proto_rev = proto_rev
            self.m_do_reconnect = True
    
            self.m_reconnect_cv.notify()




#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# When running, this will continue incrementing the value 'dsTime', and the
# value should be visible to other networktables clients and the robot.
#

import sys
import time
from networktables import NetworkTables
from networktables.util import ntproperty

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTables.initialize(server=ip)


class SomeClient(object):
    '''Demonstrates an object with magic networktables properties'''
    
    robotTime = ntproperty('/SmartDashboard/robotTime', 0, writeDefault=False)
    
    dsTime = ntproperty('/SmartDashboard/dsTime', 0)

c = SomeClient()


i = 0
while True:
    
    # equivalent to wpilib.SmartDashboard.getNumber('robotTime')
    print('robotTime:', c.robotTime)
    
    # equivalent to wpilib.SmartDashboard.putNumber('dsTime', i)
    c.dsTime = i
    
    time.sleep(1)
    i += 1

